from functools import partial, wraps

from aiohttp import web

from . import models
from .backends.base import PublicException


def _wrap_error_handler(handler):
    @wraps(handler)
    async def inner(request):
        try:
            return await handler(request)
        except web.HTTPException as exc:
            headers = exc.headers
            headers.popall("Content-Type", None)
            return web.json_response(
                {"error": exc.reason}, headers=headers, status=exc.status
            )

    return inner


def _wrap_authenticated(handler, user=False, token=False):
    if not user and not token:
        return handler

    @wraps(handler)
    async def inner(request):
        if token:
            backend = request.app["backend"]
            auth_header = request.headers.get("Authorization")
            if auth_header:
                auth_type, auth_token = auth_header.split(" ", 1)
                if auth_type == "token":
                    cluster_name = request.match_info["cluster_name"]
                    cluster = await backend.get_cluster(cluster_name)
                    if cluster is not None and cluster.token == auth_token:
                        # Use `none` to indicate api token authenticated
                        request["user"] = None
                        return await handler(request)
            if not user:
                raise web.HTTPUnauthorized(headers={"WWW-Authenticate": "token"})

        if user:
            auth = request.app["authenticator"]
            return await auth.authenticate_and_handle(request, handler)

    return inner


def api_handler(handler=None, user_authenticated=False, token_authenticated=False):
    if handler is None:
        return partial(
            api_handler,
            user_authenticated=user_authenticated,
            token_authenticated=token_authenticated,
        )

    handler = _wrap_authenticated(
        handler, user=user_authenticated, token=token_authenticated
    )

    return _wrap_error_handler(handler)


default_routes = web.RouteTableDef()


@default_routes.get("/api/health")
@api_handler
async def health(request):
    health = await request.app["gateway"].health()
    if health["status"] == "fail":
        status = 503
    else:
        status = 200
    return web.json_response(health, status=status)


@default_routes.get("/api/version")
@api_handler
async def version(request):
    version = request.app["gateway"].version_info()
    return web.json_response(version, status=200)


@default_routes.get("/api/v1/options")
@api_handler(user_authenticated=True)
async def cluster_options(request):
    user = request["user"]
    backend = request.app["backend"]
    cluster_options = await backend.get_cluster_options(user)
    spec = cluster_options.get_specification()
    return web.json_response({"cluster_options": spec})


@default_routes.get("/api/v1/clusters/")
@api_handler(user_authenticated=True)
async def list_clusters(request):
    user = request["user"]
    backend = request.app["backend"]
    status = request.query.get("status")
    if status is None:
        statuses = None
    else:
        try:
            statuses = [models.ClusterStatus.from_name(k) for k in status.split(",")]
        except Exception as exc:
            raise web.HTTPUnprocessableEntity(reason=str(exc))
    clusters = await backend.list_clusters(username=user.name, statuses=statuses)
    return web.json_response({c.name: c.to_dict() for c in clusters})


@default_routes.post("/api/v1/clusters/")
@api_handler(user_authenticated=True)
async def create_cluster(request):
    user = request["user"]
    backend = request.app["backend"]
    msg = await request.json()
    cluster_options = msg.get("cluster_options") or {}

    try:
        cluster_name = await backend.start_cluster(user, cluster_options)
    except PublicException as exc:
        reason = str(exc)
        request.app["log"].info(
            "Error creating new cluster for user %s: %s", user.name, reason
        )
        raise web.HTTPUnprocessableEntity(reason=reason)
    return web.json_response({"name": cluster_name}, status=201)


def _parse_query_flag(val):
    if val is None:
        return False
    elif val == "":
        return True
    else:
        try:
            return bool(int(val))
        except Exception:
            return False


@default_routes.get("/api/v1/clusters/{cluster_name}")
@api_handler(user_authenticated=True)
async def get_cluster(request):
    user = request["user"]
    cluster_name = request.match_info["cluster_name"]
    backend = request.app["backend"]
    wait = request.query.get("wait")
    wait = _parse_query_flag(wait)
    cluster = await backend.get_cluster(cluster_name, wait=wait)
    if cluster is None:
        raise web.HTTPNotFound(reason=f"Cluster {cluster_name} not found")
    if not user.has_permissions(cluster):
        raise web.HTTPForbidden(
            reason=f"User {user.name} lacks permissions to view cluster {cluster_name}"
        )
    return web.json_response(cluster.to_dict())


@default_routes.delete("/api/v1/clusters/{cluster_name}")
@api_handler(user_authenticated=True, token_authenticated=True)
async def delete_cluster(request):
    user = request["user"]
    cluster_name = request.match_info["cluster_name"]
    backend = request.app["backend"]
    cluster = await backend.get_cluster(cluster_name)
    if cluster is not None:
        if user is not None and not user.has_permissions(cluster):
            raise web.HTTPForbidden(
                reason=f"User {user.name} lacks permissions to stop cluster {cluster_name}"
            )
        await backend.stop_cluster(cluster_name)
    return web.Response(status=204)


@default_routes.post("/api/v1/clusters/{cluster_name}/scale")
@api_handler(user_authenticated=True)
async def scale_cluster(request):
    user = request["user"]
    cluster_name = request.match_info["cluster_name"]
    backend = request.app["backend"]
    cluster = await backend.get_cluster(cluster_name)
    if cluster is None:
        raise web.HTTPNotFound(reason=f"Cluster {cluster_name} not found")
    if not user.has_permissions(cluster):
        raise web.HTTPForbidden(
            reason=f"User {user.name} lacks permissions to view cluster {cluster_name}"
        )

    msg = await request.json()
    count = msg.get("count", None)
    if not isinstance(count, int) or count < 0:
        raise web.HTTPUnprocessableEntity(
            reason=f"Scale expects a non-negative integer, got {count}"
        )

    max_workers = cluster.config.get("cluster_max_workers")
    resp_msg = None
    if max_workers is not None and count > max_workers:
        resp_msg = (
            f"Scale request of {count} workers would exceed resource limit of "
            f"{max_workers} workers. Scaling to {max_workers} instead."
        )
        count = max_workers

    try:
        await backend.forward_message_to_scheduler(
            cluster, {"op": "scale", "count": count}
        )
    except PublicException as exc:
        raise web.HTTPConflict(reason=str(exc))

    return web.json_response({"ok": not resp_msg, "msg": resp_msg})


@default_routes.post("/api/v1/clusters/{cluster_name}/adapt")
@api_handler(user_authenticated=True)
async def adapt_cluster(request):
    user = request["user"]
    cluster_name = request.match_info["cluster_name"]
    backend = request.app["backend"]
    cluster = await backend.get_cluster(cluster_name)
    if cluster is None:
        raise web.HTTPNotFound(reason=f"Cluster {cluster_name} not found")
    if not user.has_permissions(cluster):
        raise web.HTTPForbidden(
            reason=f"User {user.name} lacks permissions to view cluster {cluster_name}"
        )

    msg = await request.json()
    minimum = msg.get("minimum", None)
    maximum = msg.get("maximum", None)
    active = msg.get("active", True)

    max_workers = cluster.config.get("cluster_max_workers")
    resp_msg = None
    if max_workers is not None:
        if maximum is None:
            maximum = max_workers
        if minimum is None:
            minimum = 0
        if maximum > max_workers or minimum > max_workers:
            orig_max = maximum
            orig_min = minimum
            maximum = min(max_workers, maximum)
            minimum = min(max_workers, minimum)
            resp_msg = (
                f"Adapt with `maximum={orig_max}, minimum={orig_min}` workers "
                f"would exceed resource limit of {max_workers} workers. Using "
                f"`maximum={maximum}, minimum={minimum}` instead."
            )

    try:
        await backend.forward_message_to_scheduler(
            cluster,
            {"op": "adapt", "minimum": minimum, "maximum": maximum, "active": active},
        )
    except PublicException as exc:
        raise web.HTTPConflict(reason=str(exc))

    return web.json_response({"ok": not resp_msg, "msg": resp_msg})


@default_routes.post("/api/v1/clusters/{cluster_name}/heartbeat")
@api_handler(token_authenticated=True)
async def handle_heartbeat(request):
    backend = request.app["backend"]
    cluster_name = request.match_info["cluster_name"]
    msg = await request.json()
    try:
        await backend.on_cluster_heartbeat(cluster_name, msg)
    except PublicException as exc:
        raise web.HTTPConflict(reason=str(exc))
    return web.Response()



from dask_gateway.scheduler_preload import GatewaySchedulerService, make_gateway_client
from distributed import Scheduler, Security, Worker
from distributed.core import Status
from distributed.utils import TimeoutError

from .local import UnsafeLocalBackend

__all__ = ("InProcessBackend",)


class InProcessBackend(UnsafeLocalBackend):
    """A backend that runs everything in the same process"""

    def get_security(self, cluster):
        cert_path, key_path = self.get_tls_paths(cluster)
        return Security(
            tls_ca_file=cert_path,
            tls_scheduler_cert=cert_path,
            tls_scheduler_key=key_path,
            tls_worker_cert=cert_path,
            tls_worker_key=key_path,
        )

    def get_gateway_client(self, cluster):
        return make_gateway_client(
            cluster_name=cluster.name, api_token=cluster.token, api_url=self.api_url
        )

    def _check_status(self, objs, mapping):
        out = []
        for x in objs:
            x = mapping.get(x.name)
            ok = x is not None and not x.status != Status.closed
            out.append(ok)
        return out

    async def do_setup(self):
        self.schedulers = {}
        self.workers = {}

    async def do_start_cluster(self, cluster):
        workdir = self.setup_working_directory(cluster)
        yield {"workdir": workdir}

        security = self.get_security(cluster)
        gateway_client = self.get_gateway_client(cluster)

        self.schedulers[cluster.name] = scheduler = Scheduler(
            protocol="tls",
            host="127.0.0.1",
            port=0,
            dashboard_address="127.0.0.1:0",
            security=security,
            services={
                ("gateway", ":0"): (
                    GatewaySchedulerService,
                    {
                        "gateway": gateway_client,
                        "heartbeat_period": self.cluster_heartbeat_period,
                        "adaptive_period": cluster.config.adaptive_period,
                        "idle_timeout": cluster.config.idle_timeout,
                    },
                )
            },
        )
        await scheduler
        yield {"workdir": workdir, "started": True}

    async def do_stop_cluster(self, cluster):
        scheduler = self.schedulers.pop(cluster.name)

        await scheduler.close()
        scheduler.stop()

        workdir = cluster.state.get("workdir")
        if workdir is not None:
            self.cleanup_working_directory(workdir)

    async def do_check_clusters(self, clusters):
        return self._check_status(clusters, self.schedulers)

    async def do_start_worker(self, worker):
        security = self.get_security(worker.cluster)
        workdir = worker.cluster.state["workdir"]
        self.workers[worker.name] = worker = Worker(
            worker.cluster.scheduler_address,
            nthreads=worker.cluster.config.worker_threads,
            memory_limit=0,
            security=security,
            name=worker.name,
            local_directory=workdir,
        )
        await worker
        yield {"started": True}

    async def do_stop_worker(self, worker):
        worker = self.workers.pop(worker.name, None)
        if worker is None:
            return
        try:
            await worker.close(timeout=1)
        except TimeoutError:
            pass

    async def do_check_workers(self, workers):
        return self._check_status(workers, self.workers)

    async def worker_status(self, worker_name, worker_state, cluster_state):
        worker = self.workers.get(worker_name)
        if worker is None:
            return False
        return not worker.status != Status.closed



import asyncio
import base64
import json
import os
import sys
import uuid
from collections import defaultdict
from itertools import chain, islice

import sqlalchemy as sa
from cryptography.fernet import Fernet, MultiFernet
from traitlets import Bool, Float, Integer, List, Unicode, default, validate

from .. import models
from ..proxy import Proxy
from ..tls import new_keypair
from ..utils import Flag, FrozenAttrDict, TaskPool, normalize_address, timestamp
from ..workqueue import Backoff, WorkQueue, WorkQueueClosed
from .base import Backend

if sys.version_info >= (3, 11):
    from asyncio import timeout
else:
    from async_timeout import timeout

__all__ = ("DBBackendBase", "Cluster", "Worker")


def _normalize_encrypt_key(key):
    if isinstance(key, str):
        key = key.encode("ascii")

    if len(key) == 44:
        try:
            key = base64.urlsafe_b64decode(key)
        except ValueError:
            pass

    if len(key) == 32:
        return base64.urlsafe_b64encode(key)

    raise ValueError(
        "All keys in `db_encrypt_keys`/`DASK_GATEWAY_ENCRYPT_KEYS` must be 32 "
        "bytes, base64-encoded"
    )


def _is_in_memory_db(url):
    return url in ("sqlite://", "sqlite:///:memory:")


class _IntEnum(sa.TypeDecorator):
    impl = sa.Integer

    def __init__(self, enumclass, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumclass = enumclass

    def process_bind_param(self, value, dialect):
        return value.value

    def process_result_value(self, value, dialect):
        return self._enumclass(value)


class _JSON(sa.TypeDecorator):
    "Represents an immutable structure as a json-encoded string."

    impl = sa.LargeBinary

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value).encode("utf-8")
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class JobStatus(models.IntEnum):
    CREATED = 1
    SUBMITTED = 2
    RUNNING = 3
    CLOSING = 4
    STOPPED = 5
    FAILED = 6


class Cluster:
    """Information on a cluster.

    Not all attributes on this object are publically accessible. When writing a
    backend, you may access the following attributes:

    Attributes
    ----------
    name : str
        The cluster name.
    username : str
        The user associated with this cluster.
    token : str
        The API token associated with this cluster. Used to authenticate the
        cluster with the gateway.
    config : FrozenAttrDict
        The serialized ``ClusterConfig`` associated with this cluster.
    state : dict
        Any cluster state, as yielded from ``do_start_cluster``.
    scheduler_address : str
        The scheduler address. The empty string if the cluster is not running.
    dashboard_address : str
        The dashboard address. The empty string if the cluster is not running,
        or no dashboard is running on the cluster.
    api_address : str
        The cluster's api address. The empty string if the cluster is not running.
    tls_cert : bytes
        The TLS cert credentials associated with the cluster.
    tls_key : bytes
        The TLS key credentials associated with the cluster.
    """

    def __init__(
        self,
        id=None,
        name=None,
        username=None,
        token=None,
        options=None,
        config=None,
        status=None,
        target=None,
        count=0,
        state=None,
        scheduler_address="",
        dashboard_address="",
        api_address="",
        tls_cert=b"",
        tls_key=b"",
        start_time=None,
        stop_time=None,
    ):
        self.id = id
        self.name = name
        self.username = username
        self.token = token
        self.options = options
        self.config = config
        self.status = status
        self.target = target
        self.count = count
        self.state = state
        self.scheduler_address = scheduler_address
        self.dashboard_address = dashboard_address
        self.api_address = api_address
        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.start_time = start_time
        self.stop_time = stop_time

        if self.status == JobStatus.RUNNING:
            self.last_heartbeat = timestamp()
        else:
            self.last_heartbeat = None
        self.worker_start_failure_count = 0
        self.added_to_proxies = False
        self.workers = {}

        self.ready = Flag()
        if self.status >= JobStatus.RUNNING:
            self.ready.set()
        self.shutdown = Flag()
        if self.status >= JobStatus.STOPPED:
            self.shutdown.set()

    _status_map = {
        (JobStatus.CREATED, JobStatus.RUNNING): models.ClusterStatus.PENDING,
        (JobStatus.CREATED, JobStatus.CLOSING): models.ClusterStatus.STOPPING,
        (JobStatus.CREATED, JobStatus.STOPPED): models.ClusterStatus.STOPPING,
        (JobStatus.CREATED, JobStatus.FAILED): models.ClusterStatus.STOPPING,
        (JobStatus.SUBMITTED, JobStatus.RUNNING): models.ClusterStatus.PENDING,
        (JobStatus.SUBMITTED, JobStatus.CLOSING): models.ClusterStatus.STOPPING,
        (JobStatus.SUBMITTED, JobStatus.STOPPED): models.ClusterStatus.STOPPING,
        (JobStatus.SUBMITTED, JobStatus.FAILED): models.ClusterStatus.STOPPING,
        (JobStatus.RUNNING, JobStatus.RUNNING): models.ClusterStatus.RUNNING,
        (JobStatus.RUNNING, JobStatus.CLOSING): models.ClusterStatus.STOPPING,
        (JobStatus.RUNNING, JobStatus.STOPPED): models.ClusterStatus.STOPPING,
        (JobStatus.RUNNING, JobStatus.FAILED): models.ClusterStatus.STOPPING,
        (JobStatus.CLOSING, JobStatus.STOPPED): models.ClusterStatus.STOPPING,
        (JobStatus.CLOSING, JobStatus.FAILED): models.ClusterStatus.STOPPING,
        (JobStatus.STOPPED, JobStatus.STOPPED): models.ClusterStatus.STOPPED,
        (JobStatus.FAILED, JobStatus.FAILED): models.ClusterStatus.FAILED,
    }

    def active_workers(self):
        return [w for w in self.workers.values() if w.is_active()]

    def is_active(self):
        return self.target < JobStatus.STOPPED

    def all_workers_at_least(self, status):
        return all(w.status >= status for w in self.workers.values())

    @property
    def model_status(self):
        return self._status_map[self.status, self.target]

    def to_model(self):
        return models.Cluster(
            name=self.name,
            username=self.username,
            token=self.token,
            options=self.options,
            config=self.config,
            status=self.model_status,
            scheduler_address=self.scheduler_address,
            dashboard_address=self.dashboard_address,
            api_address=self.api_address,
            tls_cert=self.tls_cert,
            tls_key=self.tls_key,
            start_time=self.start_time,
            stop_time=self.stop_time,
        )


class Worker:
    """Information on a worker.

    Not all attributes on this object are publicly accessible. When writing a
    backend, you may access the following attributes:

    Attributes
    ----------
    name : str
        The worker name.
    cluster : Cluster
        The cluster associated with this worker.
    state : dict
        Any worker state, as yielded from ``do_start_worker``.
    """

    def __init__(
        self,
        id=None,
        name=None,
        cluster=None,
        status=None,
        target=None,
        state=None,
        start_time=None,
        stop_time=None,
        close_expected=False,
    ):
        self.id = id
        self.name = name
        self.cluster = cluster
        self.status = status
        self.target = target
        self.state = state
        self.start_time = start_time
        self.stop_time = stop_time
        self.close_expected = close_expected

    def is_active(self):
        return self.target < JobStatus.STOPPED


metadata = sa.MetaData()

clusters = sa.Table(
    "clusters",
    metadata,
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("name", sa.Unicode(255), nullable=False, unique=True),
    sa.Column("username", sa.Unicode(255), nullable=False),
    sa.Column("status", _IntEnum(JobStatus), nullable=False),
    sa.Column("target", _IntEnum(JobStatus), nullable=False),
    sa.Column("count", sa.Integer, nullable=False),
    sa.Column("options", _JSON, nullable=False),
    sa.Column("config", _JSON, nullable=False),
    sa.Column("state", _JSON, nullable=False),
    sa.Column("token", sa.BINARY(140), nullable=False, unique=True),
    sa.Column("scheduler_address", sa.Unicode(255), nullable=False),
    sa.Column("dashboard_address", sa.Unicode(255), nullable=False),
    sa.Column("api_address", sa.Unicode(255), nullable=False),
    sa.Column("tls_credentials", sa.LargeBinary, nullable=False),
    sa.Column("start_time", sa.Integer, nullable=False),
    sa.Column("stop_time", sa.Integer, nullable=True),
)

workers = sa.Table(
    "workers",
    metadata,
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("name", sa.Unicode(255), nullable=False),
    sa.Column(
        "cluster_id", sa.ForeignKey("clusters.id", ondelete="CASCADE"), nullable=False
    ),
    sa.Column("status", _IntEnum(JobStatus), nullable=False),
    sa.Column("target", _IntEnum(JobStatus), nullable=False),
    sa.Column("state", _JSON, nullable=False),
    sa.Column("start_time", sa.Integer, nullable=False),
    sa.Column("stop_time", sa.Integer, nullable=True),
    sa.Column("close_expected", sa.Integer, nullable=False),
)


class DataManager:
    """Holds the internal state for a single Dask Gateway.

    Keeps the memory representation in-sync with the database.
    """

    def __init__(self, url="sqlite:///:memory:", encrypt_keys=(), **kwargs):
        if url.startswith("sqlite"):
            kwargs["connect_args"] = {"check_same_thread": False}

        if _is_in_memory_db(url):
            kwargs["poolclass"] = sa.pool.StaticPool
            self.fernet = None
        else:
            self.fernet = MultiFernet([Fernet(key) for key in encrypt_keys])

        engine = sa.create_engine(url, **kwargs)
        if url.startswith("sqlite"):
            # Register PRAGMA foreigh_keys=on for sqlite
            @sa.event.listens_for(engine, "connect")
            def connect(dbapi_con, con_record):
                cursor = dbapi_con.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()

        metadata.create_all(engine)

        self.db = engine

        self.username_to_clusters = defaultdict(dict)
        self.name_to_cluster = {}
        self.id_to_cluster = {}

        # Load all existing clusters into memory
        with self.db.begin() as connection:
            for c in connection.execute(clusters.select()):
                tls_cert, tls_key = self.decode_tls_credentials(c.tls_credentials)
                token = self.decode_token(c.token)
                cluster = Cluster(
                    id=c.id,
                    name=c.name,
                    username=c.username,
                    token=token,
                    options=c.options,
                    config=FrozenAttrDict(c.config),
                    status=c.status,
                    target=c.target,
                    count=c.count,
                    state=c.state,
                    scheduler_address=c.scheduler_address,
                    dashboard_address=c.dashboard_address,
                    api_address=c.api_address,
                    tls_cert=tls_cert,
                    tls_key=tls_key,
                    start_time=c.start_time,
                    stop_time=c.stop_time,
                )
                self.username_to_clusters[cluster.username][cluster.name] = cluster
                self.id_to_cluster[cluster.id] = cluster
                self.name_to_cluster[cluster.name] = cluster

            # Next load all existing workers into memory
            for w in connection.execute(workers.select()):
                cluster = self.id_to_cluster[w.cluster_id]
                worker = Worker(
                    id=w.id,
                    name=w.name,
                    status=w.status,
                    target=w.target,
                    cluster=cluster,
                    state=w.state,
                    start_time=w.start_time,
                    stop_time=w.stop_time,
                    close_expected=w.close_expected,
                )
                cluster.workers[worker.name] = worker

    def cleanup_expired(self, max_age_in_seconds):
        cutoff = timestamp() - max_age_in_seconds * 1000
        with self.db.begin() as conn:
            to_delete = conn.execute(
                sa.select(clusters.c.id).where(clusters.c.stop_time < cutoff)
            ).fetchall()

            if to_delete:
                to_delete = [i for i, in to_delete]

                conn.execute(
                    clusters.delete().where(clusters.c.id == sa.bindparam("id")),
                    [{"id": i} for i in to_delete],
                )

                for i in to_delete:
                    cluster = self.id_to_cluster.pop(i)
                    self.name_to_cluster.pop(cluster.name, None)
                    user_clusters = self.username_to_clusters[cluster.username]
                    user_clusters.pop(cluster.name)
                    if not user_clusters:
                        self.username_to_clusters.pop(cluster.username)

        return len(to_delete)

    def encrypt(self, b):
        """Encrypt bytes ``b``. If encryption is disabled this is a no-op"""
        return b if self.fernet is None else self.fernet.encrypt(b)

    def decrypt(self, b):
        """Decrypt bytes ``b``. If encryption is disabled this is a no-op"""
        return b if self.fernet is None else self.fernet.decrypt(b)

    def encode_tls_credentials(self, tls_cert, tls_key):
        return self.encrypt(b";".join((tls_cert, tls_key)))

    def decode_tls_credentials(self, data):
        return self.decrypt(data).split(b";")

    def encode_token(self, token):
        return self.encrypt(token.encode("utf8"))

    def decode_token(self, data):
        return self.decrypt(data).decode()

    def get_cluster(self, cluster_name):
        return self.name_to_cluster.get(cluster_name)

    def list_clusters(self, username=None, statuses=None):
        if statuses is None:
            select = lambda x: x.is_active()
        else:
            statuses = set(statuses)
            select = lambda x: x.model_status in statuses
        if username is None:
            return [
                cluster for cluster in self.name_to_cluster.values() if select(cluster)
            ]
        else:
            clusters = self.username_to_clusters.get(username)
            if clusters is None:
                return []
            return [cluster for cluster in clusters.values() if select(cluster)]

    def active_clusters(self):
        for cluster in self.name_to_cluster.values():
            if cluster.is_active():
                yield cluster

    def create_cluster(self, username, options, config):
        """Create a new cluster for a user"""
        cluster_name = uuid.uuid4().hex
        token = uuid.uuid4().hex
        tls_cert, tls_key = new_keypair(cluster_name)
        # Encode the tls credentials for storing in the database
        tls_credentials = self.encode_tls_credentials(tls_cert, tls_key)
        enc_token = self.encode_token(token)

        common = {
            "name": cluster_name,
            "username": username,
            "options": options,
            "status": JobStatus.CREATED,
            "target": JobStatus.RUNNING,
            "count": 0,
            "state": {},
            "scheduler_address": "",
            "dashboard_address": "",
            "api_address": "",
            "start_time": timestamp(),
        }

        with self.db.begin() as conn:
            res = conn.execute(
                clusters.insert().values(
                    tls_credentials=tls_credentials,
                    token=enc_token,
                    config=config,
                    **common,
                )
            )
            cluster = Cluster(
                id=res.inserted_primary_key[0],
                token=token,
                tls_cert=tls_cert,
                tls_key=tls_key,
                config=FrozenAttrDict(config),
                **common,
            )
            self.id_to_cluster[cluster.id] = cluster
            self.name_to_cluster[cluster_name] = cluster
            self.username_to_clusters[username][cluster_name] = cluster

        return cluster

    def create_worker(self, cluster):
        """Create a new worker for a cluster"""
        worker_name = uuid.uuid4().hex

        common = {
            "name": worker_name,
            "status": JobStatus.CREATED,
            "target": JobStatus.RUNNING,
            "state": {},
            "start_time": timestamp(),
            "close_expected": False,
        }

        with self.db.begin() as conn:
            res = conn.execute(workers.insert().values(cluster_id=cluster.id, **common))
            worker = Worker(id=res.inserted_primary_key[0], cluster=cluster, **common)
            cluster.workers[worker.name] = worker

        return worker

    def update_cluster(self, cluster, **kwargs):
        """Update a cluster's state"""
        with self.db.begin() as conn:
            conn.execute(
                clusters.update().where(clusters.c.id == cluster.id).values(**kwargs)
            )
            for k, v in kwargs.items():
                setattr(cluster, k, v)

    def update_clusters(self, updates):
        """Update multiple clusters' states"""
        if not updates:
            return
        with self.db.begin() as conn:
            conn.execute(
                clusters.update().where(clusters.c.id == sa.bindparam("_id")),
                [{"_id": c.id, **u} for c, u in updates],
            )
            for c, u in updates:
                for k, v in u.items():
                    setattr(c, k, v)

    def update_worker(self, worker, **kwargs):
        """Update a worker's state"""
        with self.db.begin() as conn:
            conn.execute(
                workers.update().where(workers.c.id == worker.id).values(**kwargs)
            )
            for k, v in kwargs.items():
                setattr(worker, k, v)

    def update_workers(self, updates):
        """Update multiple workers' states"""
        if not updates:
            return
        with self.db.begin() as conn:
            conn.execute(
                workers.update().where(workers.c.id == sa.bindparam("_id")),
                [{"_id": w.id, **u} for w, u in updates],
            )
            for w, u in updates:
                for k, v in u.items():
                    setattr(w, k, v)


class DBBackendBase(Backend):
    """A base class for defining backends that rely on a database for managing state.

    Subclasses should define the following methods:

    - ``do_setup``
    - ``do_cleanup``
    - ``do_start_cluster``
    - ``do_stop_cluster``
    - ``do_check_clusters``
    - ``do_start_worker``
    - ``do_stop_worker``
    - ``do_check_workers``
    """

    db_url = Unicode(
        "sqlite:///:memory:",
        help="""
        The URL for the database. Default is in-memory only.

        If not in-memory, ``db_encrypt_keys`` must also be set.
        """,
        config=True,
    )

    db_encrypt_keys = List(
        help="""
        A list of keys to use to encrypt private data in the database. Can also
        be set by the environment variable ``DASK_GATEWAY_ENCRYPT_KEYS``, where
        the value is a ``;`` delimited string of encryption keys.

        Each key should be a base64-encoded 32 byte value, and should be
        cryptographically random. Lacking other options, openssl can be used to
        generate a single key via:

        .. code-block:: shell

            $ openssl rand -base64 32

        A single key is valid, multiple keys can be used to support key rotation.
        """,
        config=True,
    )

    @default("db_encrypt_keys")
    def _db_encrypt_keys_default(self):
        keys = [
            k.strip()
            for k in os.environb.get(b"DASK_GATEWAY_ENCRYPT_KEYS", b"").split(b";")
            if k.strip()
        ]
        return self._db_encrypt_keys_validate({"value": keys})

    @validate("db_encrypt_keys")
    def _db_encrypt_keys_validate(self, proposal):
        if not proposal["value"] and not _is_in_memory_db(self.db_url):
            raise ValueError(
                "Must configure `db_encrypt_keys`/`DASK_GATEWAY_ENCRYPT_KEYS` "
                "when not using an in-memory database"
            )
        return [_normalize_encrypt_key(k) for k in proposal["value"]]

    db_debug = Bool(
        False, help="If True, all database operations will be logged", config=True
    )

    db_cleanup_period = Float(
        600,
        help="""
        Time (in seconds) between database cleanup tasks.

        This sets how frequently old records are removed from the database.
        This shouldn't be too small (to keep the overhead low), but should be
        smaller than ``db_record_max_age`` (probably by an order of magnitude).
        """,
        config=True,
    )

    db_cluster_max_age = Float(
        3600 * 24,
        help="""
        Max time (in seconds) to keep around records of completed clusters.

        Every ``db_cleanup_period``, completed clusters older than
        ``db_cluster_max_age`` are removed from the database.
        """,
        config=True,
    )

    stop_clusters_on_shutdown = Bool(
        True,
        help="""
        Whether to stop active clusters on gateway shutdown.

        If true, all active clusters will be stopped before shutting down the
        gateway. Set to False to leave active clusters running.
        """,
        config=True,
    )

    @validate("stop_clusters_on_shutdown")
    def _stop_clusters_on_shutdown_validate(self, proposal):
        if not proposal.value and _is_in_memory_db(self.db_url):
            raise ValueError(
                "When using an in-memory database, `stop_clusters_on_shutdown` "
                "must be True"
            )
        return proposal.value

    cluster_status_period = Float(
        30,
        help="""
        Time (in seconds) between cluster status checks.

        A smaller period will detect failed clusters sooner, but will use more
        resources. A larger period will provide slower feedback in the presence
        of failures.
        """,
        config=True,
    )

    worker_status_period = Float(
        30,
        help="""
        Time (in seconds) between worker status checks.

        A smaller period will detect failed workers sooner, but will use more
        resources. A larger period will provide slower feedback in the presence
        of failures.
        """,
        config=True,
    )

    cluster_heartbeat_period = Integer(
        15,
        help="""
        Time (in seconds) between cluster heartbeats to the gateway.

        A smaller period will detect failed workers sooner, but will use more
        resources. A larger period will provide slower feedback in the presence
        of failures.
        """,
        config=True,
    )

    cluster_heartbeat_timeout = Float(
        help="""
        Timeout (in seconds) before killing a dask cluster that's failed to heartbeat.

        This should be greater than ``cluster_heartbeat_period``.  Defaults to
        ``2 * cluster_heartbeat_period``.
        """,
        config=True,
    )

    @default("cluster_heartbeat_timeout")
    def _default_cluster_heartbeat_timeout(self):
        return self.cluster_heartbeat_period * 2

    cluster_start_timeout = Float(
        60,
        help="""
        Timeout (in seconds) before giving up on a starting dask cluster.
        """,
        config=True,
    )

    worker_start_timeout = Float(
        60,
        help="""
        Timeout (in seconds) before giving up on a starting dask worker.
        """,
        config=True,
    )

    check_timeouts_period = Float(
        help="""
        Time (in seconds) between timeout checks.

        This shouldn't be too small (to keep the overhead low), but should be
        smaller than ``cluster_heartbeat_timeout``, ``cluster_start_timeout``,
        and ``worker_start_timeout``.
        """,
        config=True,
    )

    @default("check_timeouts_period")
    def _default_check_timeouts_period(self):
        min_timeout = min(
            self.cluster_heartbeat_timeout,
            self.cluster_start_timeout,
            self.worker_start_timeout,
        )
        return min(20, min_timeout / 2)

    worker_start_failure_limit = Integer(
        3,
        help="""
        A limit on the number of failed attempts to start a worker before the
        cluster is marked as failed.

        Every worker that fails to start (timeouts exempt) increments a
        counter. The counter is reset if a worker successfully starts. If the
        counter ever exceeds this limit, the cluster is marked as failed and is
        shutdown.
        """,
        config=True,
    )

    parallelism = Integer(
        20,
        help="""
        Number of handlers to use for starting/stopping clusters.
        """,
        config=True,
    )

    backoff_base_delay = Float(
        0.1,
        help="""
        Base delay (in seconds) for backoff when retrying after failures.

        If an operation fails, it is retried after a backoff computed as:

        ```
        min(backoff_max_delay, backoff_base_delay * 2 ** num_failures)
        ```
        """,
        config=True,
    )

    backoff_max_delay = Float(
        300,
        help="""
        Max delay (in seconds) for backoff policy when retrying after failures.
        """,
        config=True,
    )

    api_url = Unicode(
        help="""
        The address that internal components (e.g. dask clusters)
        will use when contacting the gateway.

        Defaults to `{proxy_address}/{prefix}/api`, set manually if a different
        address should be used.
        """,
        config=True,
    )

    @default("api_url")
    def _api_url_default(self):
        proxy = self.proxy
        scheme = "https" if proxy.tls_cert else "http"
        address = normalize_address(proxy.address, resolve_host=True)
        return f"{scheme}://{address}{proxy.prefix}/api"

    async def setup(self, app):
        await super().setup(app)

        # Setup reconcilation queues
        self.queue = WorkQueue(
            backoff=Backoff(
                base_delay=self.backoff_base_delay, max_delay=self.backoff_max_delay
            )
        )
        self.reconcilers = [
            asyncio.ensure_future(self.reconciler_loop())
            for _ in range(self.parallelism)
        ]

        # Start the proxy
        self.proxy = Proxy(parent=self, log=self.log)
        await self.proxy.setup(app)

        # Load the database
        self.db = DataManager(
            url=self.db_url, echo=self.db_debug, encrypt_keys=self.db_encrypt_keys
        )

        # Start background tasks
        self.task_pool = TaskPool()
        self.task_pool.spawn(self.check_timeouts_loop())
        self.task_pool.spawn(self.check_clusters_loop())
        self.task_pool.spawn(self.check_workers_loop())
        self.task_pool.spawn(self.cleanup_db_loop())

        # Load all active clusters/workers into reconcilation queues
        for cluster in self.db.name_to_cluster.values():
            if cluster.status < JobStatus.STOPPED:
                self.queue.put(cluster)
                for worker in cluster.workers.values():
                    if worker.status < JobStatus.STOPPED:
                        self.queue.put(worker)

        # Further backend-specific setup
        await self.do_setup()

        self.log.info(
            "Backend started, clusters will contact api server at %s", self.api_url
        )

    async def cleanup(self):
        if hasattr(self, "task_pool"):
            # Stop background tasks
            await self.task_pool.close()

        if hasattr(self, "db"):
            if self.stop_clusters_on_shutdown:
                # Request all active clusters be stopped
                active = list(self.db.active_clusters())
                if active:
                    self.log.info("Stopping %d active clusters...", len(active))
                    self.db.update_clusters(
                        [(c, {"target": JobStatus.FAILED}) for c in active]
                    )
                    for c in active:
                        self.queue.put(c)

                # Wait until all clusters are shutdown
                pending_shutdown = [
                    c
                    for c in self.db.name_to_cluster.values()
                    if c.status < JobStatus.STOPPED
                ]
                if pending_shutdown:
                    await asyncio.wait(
                        [asyncio.ensure_future(c.shutdown) for c in pending_shutdown]
                    )

        # Stop reconcilation queues
        if hasattr(self, "reconcilers"):
            self.queue.close()
            await asyncio.gather(*self.reconcilers, return_exceptions=True)

        await self.do_cleanup()

        if hasattr(self, "proxy"):
            await self.proxy.cleanup()

        await super().cleanup()

    async def list_clusters(self, username=None, statuses=None):
        clusters = self.db.list_clusters(username=username, statuses=statuses)
        return [c.to_model() for c in clusters]

    async def get_cluster(self, cluster_name, wait=False):
        cluster = self.db.get_cluster(cluster_name)
        if cluster is None:
            return None
        if wait:
            try:
                await asyncio.wait_for(cluster.ready, 20)
            except asyncio.TimeoutError:
                pass
        return cluster.to_model()

    async def start_cluster(self, user, cluster_options):
        options, config = await self.process_cluster_options(user, cluster_options)
        cluster = self.db.create_cluster(user.name, options, config.to_dict())
        self.log.info("Created cluster %s for user %s", cluster.name, user.name)
        self.queue.put(cluster)
        return cluster.name

    async def stop_cluster(self, cluster_name, failed=False):
        cluster = self.db.get_cluster(cluster_name)
        if cluster is None:
            return
        if cluster.target <= JobStatus.RUNNING:
            self.log.info("Stopping cluster %s", cluster.name)
            target = JobStatus.FAILED if failed else JobStatus.STOPPED
            self.db.update_cluster(cluster, target=target)
            self.queue.put(cluster)

    async def on_cluster_heartbeat(self, cluster_name, msg):
        cluster = self.db.get_cluster(cluster_name)
        if cluster is None or cluster.target > JobStatus.RUNNING:
            return

        cluster.last_heartbeat = timestamp()

        if cluster.status == JobStatus.RUNNING:
            cluster_update = {}
        else:
            cluster_update = {
                "api_address": msg["api_address"],
                "scheduler_address": msg["scheduler_address"],
                "dashboard_address": msg["dashboard_address"],
            }

        count = msg["count"]
        active_workers = set(msg["active_workers"])
        closing_workers = set(msg["closing_workers"])
        closed_workers = set(msg["closed_workers"])

        self.log.info(
            "Cluster %s heartbeat [count: %d, n_active: %d, n_closing: %d, n_closed: %d]",
            cluster_name,
            count,
            len(active_workers),
            len(closing_workers),
            len(closed_workers),
        )

        max_workers = cluster.config.get("cluster_max_workers")
        if max_workers is not None and count > max_workers:
            # This shouldn't happen under normal operation, but could if the
            # user does something malicious (or there's a bug).
            self.log.info(
                "Cluster %s heartbeat requested %d workers, exceeding limit of %s.",
                cluster_name,
                count,
                max_workers,
            )
            count = max_workers

        if count != cluster.count:
            cluster_update["count"] = count

        created_workers = []
        submitted_workers = []
        target_updates = []
        newly_running = []
        close_expected = []
        for worker in cluster.workers.values():
            if worker.status >= JobStatus.STOPPED:
                continue
            elif worker.name in closing_workers:
                if worker.status < JobStatus.RUNNING:
                    newly_running.append(worker)
                close_expected.append(worker)
            elif worker.name in active_workers:
                if worker.status < JobStatus.RUNNING:
                    newly_running.append(worker)
            elif worker.name in closed_workers:
                target = (
                    JobStatus.STOPPED if worker.close_expected else JobStatus.FAILED
                )
                target_updates.append((worker, {"target": target}))
            else:
                if worker.status == JobStatus.SUBMITTED:
                    submitted_workers.append(worker)
                else:
                    assert worker.status == JobStatus.CREATED
                    created_workers.append(worker)

        n_pending = len(created_workers) + len(submitted_workers)
        n_to_stop = len(active_workers) + n_pending - count
        if n_to_stop > 0:
            for w in islice(chain(created_workers, submitted_workers), n_to_stop):
                target_updates.append((w, {"target": JobStatus.STOPPED}))

        if cluster_update:
            self.db.update_cluster(cluster, **cluster_update)
            self.queue.put(cluster)

        self.db.update_workers(target_updates)
        for w, u in target_updates:
            self.queue.put(w)

        if newly_running:
            # At least one worker successfully started, reset failure count
            cluster.worker_start_failure_count = 0
            self.db.update_workers(
                [(w, {"status": JobStatus.RUNNING}) for w in newly_running]
            )
            for w in newly_running:
                self.log.info("Worker %s is running", w.name)

        self.db.update_workers([(w, {"close_expected": True}) for w in close_expected])

    async def check_timeouts_loop(self):
        while True:
            await asyncio.sleep(self.check_timeouts_period)
            try:
                await self._check_timeouts()
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.warning(
                    "Exception while checking for timed out clusters/workers",
                    exc_info=exc,
                )

    async def _check_timeouts(self):
        self.log.debug("Checking for timed out clusters/workers")
        now = timestamp()
        cluster_heartbeat_cutoff = now - self.cluster_heartbeat_timeout * 1000
        cluster_start_cutoff = now - self.cluster_start_timeout * 1000
        worker_start_cutoff = now - self.worker_start_timeout * 1000
        cluster_updates = []
        worker_updates = []
        for cluster in self.db.active_clusters():
            if cluster.status == JobStatus.SUBMITTED:
                # Check if submitted clusters have timed out
                if cluster.start_time < cluster_start_cutoff:
                    self.log.info("Cluster %s startup timed out", cluster.name)
                    cluster_updates.append((cluster, {"target": JobStatus.FAILED}))
            elif cluster.status == JobStatus.RUNNING:
                # Check if running clusters have missed a heartbeat
                if cluster.last_heartbeat < cluster_heartbeat_cutoff:
                    self.log.info("Cluster %s heartbeat timed out", cluster.name)
                    cluster_updates.append((cluster, {"target": JobStatus.FAILED}))
                else:
                    for w in cluster.workers.values():
                        # Check if submitted workers have timed out
                        if (
                            w.status == JobStatus.SUBMITTED
                            and w.target == JobStatus.RUNNING
                            and w.start_time < worker_start_cutoff
                        ):
                            self.log.info("Worker %s startup timed out", w.name)
                            worker_updates.append((w, {"target": JobStatus.FAILED}))
        self.db.update_clusters(cluster_updates)
        for c, _ in cluster_updates:
            self.queue.put(c)
        self.db.update_workers(worker_updates)
        for w, _ in worker_updates:
            self.queue.put(w)

    async def check_clusters_loop(self):
        while True:
            await asyncio.sleep(self.cluster_status_period)
            self.log.debug("Checking pending cluster statuses")
            try:
                clusters = [
                    c
                    for c in self.db.active_clusters()
                    if c.status == JobStatus.SUBMITTED
                ]
                statuses = await self.do_check_clusters(clusters)
                updates = [
                    (c, {"target": JobStatus.FAILED})
                    for c, ok in zip(clusters, statuses)
                    if not ok
                ]
                self.db.update_clusters(updates)
                for c, _ in updates:
                    self.log.info("Cluster %s failed during startup", c.name)
                    self.queue.put(c)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.warning(
                    "Exception while checking cluster statuses", exc_info=exc
                )

    async def check_workers_loop(self):
        while True:
            await asyncio.sleep(self.worker_status_period)
            self.log.debug("Checking pending worker statuses")
            try:
                clusters = (
                    c
                    for c in self.db.active_clusters()
                    if c.status == JobStatus.RUNNING
                )
                workers = [
                    w
                    for c in clusters
                    for w in c.active_workers()
                    if w.status == JobStatus.SUBMITTED
                ]
                statuses = await self.do_check_workers(workers)
                updates = [
                    (w, {"target": JobStatus.FAILED})
                    for w, ok in zip(workers, statuses)
                    if not ok
                ]
                self.db.update_workers(updates)
                for w, _ in updates:
                    self.log.info("Worker %s failed during startup", w.name)
                    w.cluster.worker_start_failure_count += 1
                    self.queue.put(w)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.warning(
                    "Exception while checking worker statuses", exc_info=exc
                )

    async def cleanup_db_loop(self):
        while True:
            try:
                n = self.db.cleanup_expired(self.db_cluster_max_age)
            except Exception as exc:
                self.log.error(
                    "Error while cleaning expired database records", exc_info=exc
                )
            else:
                self.log.debug("Removed %d expired clusters from the database", n)
            await asyncio.sleep(self.db_cleanup_period)

    async def reconciler_loop(self):
        while True:
            try:
                obj = await self.queue.get()
            except WorkQueueClosed:
                return

            if isinstance(obj, Cluster):
                method = self.reconcile_cluster
                kind = "cluster"
            else:
                method = self.reconcile_worker
                kind = "worker"

            self.log.debug(
                "Reconciling %s %s, %s -> %s",
                kind,
                obj.name,
                obj.status.name,
                obj.target.name,
            )

            try:
                await method(obj)
            except Exception:
                self.log.warning(
                    "Error while reconciling %s %s", kind, obj.name, exc_info=True
                )
                self.queue.put_backoff(obj)
            else:
                self.queue.reset_backoff(obj)
            finally:
                self.queue.task_done(obj)

    async def reconcile_cluster(self, cluster):
        if cluster.status >= JobStatus.STOPPED:
            return

        if cluster.target in (JobStatus.STOPPED, JobStatus.FAILED):
            if cluster.status == JobStatus.CLOSING:
                if self.is_cluster_ready_to_close(cluster):
                    await self._cluster_to_stopped(cluster)
            else:
                await self._cluster_to_closing(cluster)
            return

        if cluster.target == JobStatus.RUNNING:
            if cluster.status == JobStatus.CREATED:
                await self._cluster_to_submitted(cluster)
                return

            if cluster.status == JobStatus.SUBMITTED and cluster.scheduler_address:
                await self._cluster_to_running(cluster)

            if cluster.status == JobStatus.RUNNING:
                await self._check_cluster_proxied(cluster)
                await self._check_cluster_scale(cluster)

    async def reconcile_worker(self, worker):
        if worker.status >= JobStatus.STOPPED:
            return

        if worker.target == JobStatus.CLOSING:
            if worker.status != JobStatus.CLOSING:
                self.db.update_worker(worker, status=JobStatus.CLOSING)
            if self.is_cluster_ready_to_close(worker.cluster):
                self.queue.put(worker.cluster)
            return

        if worker.target in (JobStatus.STOPPED, JobStatus.FAILED):
            await self._worker_to_stopped(worker)
            if self.is_cluster_ready_to_close(worker.cluster):
                self.queue.put(worker.cluster)
            elif (
                worker.cluster.target == JobStatus.RUNNING and not worker.close_expected
            ):
                self.queue.put(worker.cluster)
            return

        if worker.status == JobStatus.CREATED and worker.target == JobStatus.RUNNING:
            await self._worker_to_submitted(worker)
            return

    def is_cluster_ready_to_close(self, cluster):
        return (
            cluster.status == JobStatus.CLOSING
            and (
                self.supports_bulk_shutdown
                and cluster.all_workers_at_least(JobStatus.CLOSING)
            )
            or cluster.all_workers_at_least(JobStatus.STOPPED)
        )

    async def _cluster_to_submitted(self, cluster):
        self.log.info("Submitting cluster %s...", cluster.name)
        try:
            async with timeout(self.cluster_start_timeout):
                async for state in self.do_start_cluster(cluster):
                    self.log.debug("State update for cluster %s", cluster.name)
                    self.db.update_cluster(cluster, state=state)
            self.db.update_cluster(cluster, status=JobStatus.SUBMITTED)
            self.log.info("Cluster %s submitted", cluster.name)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            if isinstance(exc, asyncio.TimeoutError):
                self.log.info("Cluster %s startup timed out", cluster.name)
            else:
                self.log.warning(
                    "Failed to submit cluster %s", cluster.name, exc_info=exc
                )
            self.db.update_cluster(
                cluster, status=JobStatus.SUBMITTED, target=JobStatus.FAILED
            )
            self.queue.put(cluster)

    async def _cluster_to_closing(self, cluster):
        self.log.debug("Preparing to stop cluster %s", cluster.name)
        target = JobStatus.CLOSING if self.supports_bulk_shutdown else JobStatus.STOPPED
        workers = [w for w in cluster.workers.values() if w.target < target]
        self.db.update_workers([(w, {"target": target}) for w in workers])
        for w in workers:
            self.queue.put(w)
        self.db.update_cluster(cluster, status=JobStatus.CLOSING)
        if not workers:
            # If there are workers, the cluster will be enqueued after the last one closed
            # If there are no workers, requeue now
            self.queue.put(cluster)
        cluster.ready.set()

    async def _cluster_to_stopped(self, cluster):
        self.log.info("Stopping cluster %s...", cluster.name)
        if cluster.status > JobStatus.CREATED:
            try:
                await self.do_stop_cluster(cluster)
            except Exception as exc:
                self.log.warning(
                    "Exception while stopping cluster %s", cluster.name, exc_info=exc
                )
            await self.proxy.remove_route(kind="PATH", path=f"/clusters/{cluster.name}")
            await self.proxy.remove_route(kind="SNI", sni=cluster.name)
        self.log.info("Cluster %s stopped", cluster.name)
        self.db.update_workers(
            [
                (w, {"status": JobStatus.STOPPED, "target": JobStatus.STOPPED})
                for w in cluster.workers.values()
                if w.status < JobStatus.STOPPED
            ]
        )
        self.db.update_cluster(cluster, status=cluster.target, stop_time=timestamp())
        cluster.ready.set()
        cluster.shutdown.set()

    async def _cluster_to_running(self, cluster):
        self.log.info("Cluster %s is running", cluster.name)
        self.db.update_cluster(cluster, status=JobStatus.RUNNING)
        cluster.ready.set()

    async def _check_cluster_proxied(self, cluster):
        if not cluster.added_to_proxies:
            self.log.info("Adding cluster %s routes to proxies", cluster.name)
            if cluster.dashboard_address:
                await self.proxy.add_route(
                    kind="PATH",
                    path=f"/clusters/{cluster.name}",
                    target=cluster.dashboard_address,
                )
            await self.proxy.add_route(
                kind="SNI", sni=cluster.name, target=cluster.scheduler_address
            )
            cluster.added_to_proxies = True

    async def _check_cluster_scale(self, cluster):
        if cluster.worker_start_failure_count >= self.worker_start_failure_limit:
            self.log.info(
                "Cluster %s had %d consecutive workers fail to start, failing the cluster",
                cluster.name,
                cluster.worker_start_failure_count,
            )
            self.db.update_cluster(cluster, target=JobStatus.FAILED)
            self.queue.put(cluster)
            return

        active = cluster.active_workers()
        if cluster.count > len(active):
            for _ in range(cluster.count - len(active)):
                worker = self.db.create_worker(cluster)
                self.log.info(
                    "Created worker %s for cluster %s", worker.name, cluster.name
                )
                self.queue.put(worker)

    async def _worker_to_submitted(self, worker):
        self.log.info("Submitting worker %s...", worker.name)
        try:
            async with timeout(self.worker_start_timeout):
                async for state in self.do_start_worker(worker):
                    self.log.debug("State update for worker %s", worker.name)
                    self.db.update_worker(worker, state=state)
            self.db.update_worker(worker, status=JobStatus.SUBMITTED)
            self.log.info("Worker %s submitted", worker.name)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            if isinstance(exc, asyncio.TimeoutError):
                self.log.info("Worker %s startup timed out", worker.name)
            else:
                self.log.warning(
                    "Failed to submit worker %s", worker.name, exc_info=exc
                )
            self.db.update_worker(
                worker, status=JobStatus.SUBMITTED, target=JobStatus.FAILED
            )
            worker.cluster.worker_start_failure_count += 1
            self.queue.put(worker)

    async def _worker_to_stopped(self, worker):
        self.log.info("Stopping worker %s...", worker.name)
        if worker.status > JobStatus.CREATED:
            try:
                await self.do_stop_worker(worker)
            except Exception as exc:
                self.log.warning(
                    "Exception while stopping worker %s", worker.name, exc_info=exc
                )
        self.log.info("Worker %s stopped", worker.name)
        self.db.update_worker(worker, status=worker.target, stop_time=timestamp())

    def get_tls_paths(self, cluster):
        """Return the paths to the cert and key files for this cluster"""
        return "dask.crt", "dask.pem"

    def get_env(self, cluster):
        """Get a dict of environment variables to set for the process"""
        out = dict(cluster.config.environment)
        # Set values that dask-gateway needs to run
        out.update(
            {
                "DASK_GATEWAY_API_URL": self.api_url,
                "DASK_GATEWAY_API_TOKEN": cluster.token,
                "DASK_GATEWAY_CLUSTER_NAME": cluster.name,
                "DASK_DISTRIBUTED__COMM__REQUIRE_ENCRYPTION": "True",
            }
        )
        return out

    def get_scheduler_env(self, cluster):
        env = self.get_env(cluster)
        tls_cert_path, tls_key_path = self.get_tls_paths(cluster)
        env.update(
            {
                "DASK_DISTRIBUTED__COMM__TLS__CA_FILE": tls_cert_path,
                "DASK_DISTRIBUTED__COMM__TLS__SCHEDULER__KEY": tls_key_path,
                "DASK_DISTRIBUTED__COMM__TLS__SCHEDULER__CERT": tls_cert_path,
            }
        )
        return env

    def get_worker_env(self, cluster):
        env = self.get_env(cluster)
        tls_cert_path, tls_key_path = self.get_tls_paths(cluster)
        env.update(
            {
                "DASK_DISTRIBUTED__COMM__TLS__CA_FILE": tls_cert_path,
                "DASK_DISTRIBUTED__COMM__TLS__WORKER__KEY": tls_key_path,
                "DASK_DISTRIBUTED__COMM__TLS__WORKER__CERT": tls_cert_path,
            }
        )
        return env

    default_host = "0.0.0.0"

    def get_scheduler_command(self, cluster):
        return cluster.config.scheduler_cmd + [
            "--protocol",
            "tls",
            "--port",
            "0",
            "--host",
            self.default_host,
            "--dashboard-address",
            f"{self.default_host}:0",
            "--preload",
            "dask_gateway.scheduler_preload",
            "--dg-api-address",
            f"{self.default_host}:0",
            "--dg-heartbeat-period",
            str(self.cluster_heartbeat_period),
            "--dg-adaptive-period",
            str(cluster.config.adaptive_period),
            "--dg-idle-timeout",
            str(cluster.config.idle_timeout),
        ]

    def worker_nthreads_memory_limit_args(self, cluster):
        return str(cluster.config.worker_threads), str(cluster.config.worker_memory)

    def get_worker_command(self, cluster, worker_name, scheduler_address=None):
        nthreads, memory_limit = self.worker_nthreads_memory_limit_args(cluster)
        if scheduler_address is None:
            scheduler_address = cluster.scheduler_address
        return cluster.config.worker_cmd + [
            scheduler_address,
            "--dashboard-address",
            f"{self.default_host}:0",
            "--name",
            worker_name,
            "--nthreads",
            nthreads,
            "--memory-limit",
            memory_limit,
        ]

    # Subclasses should implement these methods
    supports_bulk_shutdown = False

    async def do_setup(self):
        """Called when the server is starting up.

        Do any initialization here.
        """
        pass

    async def do_cleanup(self):
        """Called when the server is shutting down.

        Do any cleanup here."""
        pass

    async def do_start_cluster(self, cluster):
        """Start a cluster.

        This should do any initialization for the whole dask cluster
        application, and then start the scheduler.

        Parameters
        ----------
        cluster : Cluster
            Information on the cluster to be started.

        Yields
        ------
        cluster_state : dict
            Any state needed for further interactions with this cluster. This
            should be serializable using ``json.dumps``. If startup occurs in
            multiple stages, can iteratively yield state updates to be
            checkpointed. If an error occurs at any time, the last yielded
            state will be used when calling ``do_stop_cluster``.
        """
        raise NotImplementedError

    async def do_stop_cluster(self, cluster):
        """Stop a cluster.

        Parameters
        ----------
        cluster : Cluster
            Information on the cluster to be stopped.
        """
        raise NotImplementedError

    async def do_check_clusters(self, clusters):
        """Check the status of multiple clusters.

        This is periodically called to check the status of pending clusters.
        Once a cluster is running this will no longer be called.

        Parameters
        ----------
        clusters : List[Cluster]
            The clusters to be checked.

        Returns
        -------
        statuses : List[bool]
            The status for each cluster. Return False if the cluster has
            stopped or failed, True if the cluster is pending start or running.
        """
        raise NotImplementedError

    async def do_start_worker(self, worker):
        """Start a worker.

        Parameters
        ----------
        worker : Worker
            Information on the worker to be started.

        Yields
        ------
        worker_state : dict
            Any state needed for further interactions with this worker. This
            should be serializable using ``json.dumps``. If startup occurs in
            multiple stages, can iteratively yield state updates to be
            checkpointed. If an error occurs at any time, the last yielded
            state will be used when calling ``do_stop_worker``.
        """
        raise NotImplementedError

    async def do_stop_worker(self, worker):
        """Stop a worker.

        Parameters
        ----------
        worker : Worker
            Information on the worker to be stopped.
        """
        raise NotImplementedError

    async def do_check_workers(self, workers):
        """Check the status of multiple workers.

        This is periodically called to check the status of pending workers.
        Once a worker is running this will no longer be called.

        Parameters
        ----------
        workers : List[Worker]
            The workers to be checked.

        Returns
        -------
        statuses : List[bool]
            The status for each worker. Return False if the worker has
            stopped or failed, True if the worker is pending start or running.
        """
        raise NotImplementedError



from datetime import datetime, timedelta

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


def new_keypair(sni):
    """Create a new self-signed certificate & key pair with the given SNI.

    Parameters
    ----------
    sni : str
        The SNI name to use.

    Returns
    -------
    cert_bytes : bytes
    key_bytes :  bytes
    """
    key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )
    key_bytes = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    dask_internal = x509.Name(
        [x509.NameAttribute(NameOID.COMMON_NAME, "dask-internal")]
    )
    altnames = x509.SubjectAlternativeName(
        [
            x509.DNSName(sni),
            x509.DNSName("dask-internal"),
            # allow skein appmaster and dask to share credentials
            x509.DNSName("skein-internal"),
        ]
    )
    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(dask_internal)
        .issuer_name(dask_internal)
        .add_extension(altnames, critical=False)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=365))
        .sign(key, hashes.SHA256(), default_backend())
    )

    cert_bytes = cert.public_bytes(serialization.Encoding.PEM)

    return cert_bytes, key_bytes



from traitlets import Integer, List, TraitError, TraitType
from traitlets import Type as _Type
from traitlets import Unicode
from traitlets.config import Application

# We replace the class of the default formatter used via a configuration change.
#
# References:
#
# - Traitlets' Application.logging_config defaults:
#   https://github.com/ipython/traitlets/blob/e2c731ef72dd41d4be527d4d93dd87ccc409830d/traitlets/config/application.py#L229-L256
# - Python official schema for Application.logging_config:
#   https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
#
Application.logging_config = {
    "formatters": {
        "console": {
            "class": "dask_gateway_server.utils.LogFormatter",
        },
    },
}
Application.log_level.default_value = "INFO"
Application.log_format.default_value = (
    "%(log_color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d "
    "%(name)s]%(reset)s %(message)s"
)


# Adapted from JupyterHub
class MemoryLimit(Integer):
    """A specification of a memory limit, with optional units.

    Supported units are:
      - K -> Kibibytes
      - M -> Mebibytes
      - G -> Gibibytes
      - T -> Tebibytes
    """

    UNIT_SUFFIXES = {"K": 2**10, "M": 2**20, "G": 2**30, "T": 2**40}

    def validate(self, obj, value):
        if isinstance(value, (int, float)):
            return int(value)

        try:
            num = float(value[:-1])
        except ValueError:
            raise TraitError(
                "{val} is not a valid memory specification. Must be an int or "
                "a string with suffix K, M, G, T".format(val=value)
            )
        suffix = value[-1]

        if suffix not in self.UNIT_SUFFIXES:
            raise TraitError(
                "{val} is not a valid memory specification. Must be an int or "
                "a string with suffix K, M, G, T".format(val=value)
            )
        return int(float(num) * self.UNIT_SUFFIXES[suffix])


class Callable(TraitType):
    """A trait which is callable"""

    info_text = "a callable"

    def validate(self, obj, value):
        if callable(value):
            return value
        else:
            self.error(obj, value)


class Type(_Type):
    """An implementation of `Type` with better errors"""

    def validate(self, obj, value):
        if isinstance(value, str):
            try:
                value = self._resolve_string(value)
            except ImportError as exc:
                raise TraitError(
                    "Failed to import %r for trait '%s.%s':\n\n%s"
                    % (value, type(obj).__name__, self.name, exc)
                )
        return super().validate(obj, value)


class Command(List):
    """Traitlet for a command that should be a list of strings,
    but allows it to be specified as a single string.
    """

    def __init__(self, default_value=None, **kwargs):
        kwargs.setdefault("minlen", 1)
        if isinstance(default_value, str):
            default_value = [default_value]
        super().__init__(Unicode(), default_value, **kwargs)

    def validate(self, obj, value):
        if isinstance(value, str):
            value = [value]
        return super().validate(obj, value)



from asyncio import get_running_loop
from collections import defaultdict
from tempfile import NamedTemporaryFile

try:
    import skein
except ImportError:
    raise ImportError(
        "'%s.YarnBackend' requires 'skein' as a dependency. "
        "To install required dependencies, use:\n"
        "  $ pip install dask-gateway-server[yarn]\n"
        "or\n"
        "  $ conda install dask-gateway-server-yarn -c conda-forge\n" % __name__
    )


from traitlets import Dict, Integer, Unicode

from ..traitlets import Type
from ..utils import LRUCache
from .base import ClusterConfig
from .db_base import DBBackendBase

__all__ = ("YarnClusterConfig", "YarnBackend")


class YarnClusterConfig(ClusterConfig):
    """Dask cluster configuration options when running on Hadoop/YARN"""

    queue = Unicode(
        "default", help="The YARN queue to submit applications under", config=True
    )

    localize_files = Dict(
        help="""
        Extra files to distribute to both the worker and scheduler containers.

        This is a mapping from ``local-name`` to ``resource``. Resource paths
        can be local, or in HDFS (prefix with ``hdfs://...`` if so). If an
        archive (``.tar.gz`` or ``.zip``), the resource will be unarchived as
        directory ``local-name``. For finer control, resources can also be
        specified as ``skein.File`` objects, or their ``dict`` equivalents.

        This can be used to distribute conda/virtual environments by
        configuring the following:

        .. code::

            c.YarnClusterConfig.localize_files = {
                'environment': {
                    'source': 'hdfs:///path/to/archived/environment.tar.gz',
                    'visibility': 'public'
                }
            }
            c.YarnClusterConfig.scheduler_setup = 'source environment/bin/activate'
            c.YarnClusterConfig.worker_setup = 'source environment/bin/activate'

        These archives are usually created using either ``conda-pack`` or
        ``venv-pack``. For more information on distributing files, see
        https://jcristharif.com/skein/distributing-files.html.
        """,
        config=True,
    )

    worker_setup = Unicode(
        "", help="Script to run before dask worker starts.", config=True
    )

    scheduler_setup = Unicode(
        "", help="Script to run before dask scheduler starts.", config=True
    )


class YarnBackend(DBBackendBase):
    """A cluster backend for managing dask clusters on Hadoop/YARN."""

    cluster_config_class = Type(
        "dask_gateway_server.backends.yarn.YarnClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    principal = Unicode(
        None,
        help="Kerberos principal for Dask Gateway user",
        allow_none=True,
        config=True,
    )

    keytab = Unicode(
        None,
        help="Path to kerberos keytab for Dask Gateway user",
        allow_none=True,
        config=True,
    )

    app_client_cache_max_size = Integer(
        10,
        help="""
        The max size of the cache for application clients.

        A larger cache will result in improved performance, but will also use
        more resources.
        """,
        config=True,
    )

    def async_apply(self, f, *args, **kwargs):
        return get_running_loop().run_in_executor(None, lambda: f(*args, **kwargs))

    def _get_security(self, cluster):
        return skein.Security(cert_bytes=cluster.tls_cert, key_bytes=cluster.tls_key)

    async def _get_app_client(self, cluster):
        out = self.app_client_cache.get(cluster.name)
        if out is None:
            app_id = cluster.state["app_id"]
            security = self._get_security(cluster)
            if cluster.name not in self.app_address_cache:
                # Lookup and cache the application address
                report = self.skein_client.application_report(app_id)
                if report.state != "RUNNING":  # pragma: nocover
                    raise ValueError("Application %s is not running" % app_id)
                app_address = "%s:%d" % (report.host, report.port)
                self.app_address_cache[cluster.name] = app_address
            app_address = self.app_address_cache[cluster.name]
            out = skein.ApplicationClient(app_address, app_id, security=security)
            self.app_client_cache.put(cluster.name, out)
        return out

    def worker_nthreads_memory_limit_args(self, cluster):
        return "$SKEIN_RESOURCE_VCORES", "${SKEIN_RESOURCE_MEMORY}MiB"

    def _build_specification(self, cluster, cert_path, key_path):
        files = {
            k: skein.File.from_dict(v) if isinstance(v, dict) else v
            for k, v in cluster.config.localize_files.items()
        }

        files["dask.crt"] = cert_path
        files["dask.pem"] = key_path

        scheduler_cmd = " ".join(self.get_scheduler_command(cluster))
        worker_cmd = " ".join(
            self.get_worker_command(
                cluster,
                worker_name="$DASK_GATEWAY_WORKER_NAME",
                scheduler_address="$DASK_GATEWAY_SCHEDULER_ADDRESS",
            )
        )
        scheduler_script = f"{cluster.config.scheduler_setup}\n{scheduler_cmd}"
        worker_script = f"{cluster.config.worker_setup}\n{worker_cmd}"

        master = skein.Master(
            security=self._get_security(cluster),
            resources=skein.Resources(
                memory="%d b" % cluster.config.scheduler_memory,
                vcores=cluster.config.scheduler_cores,
            ),
            files=files,
            env=self.get_scheduler_env(cluster),
            script=scheduler_script,
        )

        services = {
            "dask.worker": skein.Service(
                resources=skein.Resources(
                    memory="%d b" % cluster.config.worker_memory,
                    vcores=cluster.config.worker_cores,
                ),
                instances=0,
                max_restarts=0,
                allow_failures=True,
                files=files,
                env=self.get_worker_env(cluster),
                script=worker_script,
            )
        }

        return skein.ApplicationSpec(
            name="dask-gateway",
            queue=cluster.config.queue,
            user=cluster.username,
            master=master,
            services=services,
        )

    supports_bulk_shutdown = True

    async def do_setup(self):
        self.skein_client = await self.async_apply(
            skein.Client,
            principal=self.principal,
            keytab=self.keytab,
            security=skein.Security.new_credentials(),
        )

        self.app_client_cache = LRUCache(self.app_client_cache_max_size)
        self.app_address_cache = {}

    async def do_cleanup(self):
        self.skein_client.close()

    async def do_start_cluster(self, cluster):
        with NamedTemporaryFile() as cert_fil, NamedTemporaryFile() as key_fil:
            cert_fil.write(cluster.tls_cert)
            cert_fil.file.flush()
            key_fil.write(cluster.tls_key)
            key_fil.file.flush()
            spec = self._build_specification(cluster, cert_fil.name, key_fil.name)
            app_id = await self.async_apply(self.skein_client.submit, spec)

        yield {"app_id": app_id}

    async def do_stop_cluster(self, cluster):
        app_id = cluster.state.get("app_id")
        if app_id is None:
            return

        await self.async_apply(self.skein_client.kill_application, app_id)
        # Remove cluster from caches
        self.app_client_cache.discard(cluster.name)
        self.app_address_cache.pop(cluster.name, None)

    async def do_check_clusters(self, clusters):
        results = []
        for cluster in clusters:
            app_id = cluster.state.get("app_id")
            if app_id is None:
                return False
            report = await self.async_apply(
                self.skein_client.application_report, app_id
            )
            ok = str(report.state) not in {"FAILED", "KILLED", "FINISHED"}
            results.append(ok)
        return results

    async def do_start_worker(self, worker):
        app = await self._get_app_client(worker.cluster)
        container = await self.async_apply(
            app.add_container,
            "dask.worker",
            env={
                "DASK_GATEWAY_WORKER_NAME": worker.name,
                "DASK_GATEWAY_SCHEDULER_ADDRESS": worker.cluster.scheduler_address,
            },
        )
        yield {"container_id": container.id}

    async def do_stop_worker(self, worker):
        container_id = worker.state.get("container_id")
        if container_id is None:
            return

        app = await self._get_app_client(worker.cluster)
        try:
            await self.async_apply(app.kill_container, container_id)
        except ValueError:
            pass

    async def do_check_workers(self, workers):
        grouped = defaultdict(list)
        for w in workers:
            grouped[w.cluster].append(w)

        results = {}
        for cluster, workers in grouped.items():
            app = await self._get_app_client(cluster)
            try:
                containers = await self.async_apply(
                    app.get_containers, services=("dask.worker",)
                )
                active = {c.id for c in containers}
                results.update(
                    {w.name: w.state.get("container_id") in active for w in workers}
                )
            except Exception as exc:
                self.log.debug(
                    "Error getting worker statuses for cluster %s",
                    cluster.name,
                    exc_info=exc,
                )
                results.update({w.name: False for w in workers})

        return [results[w.name] for w in workers]



import copy
import inspect
import textwrap
from collections import OrderedDict
from collections.abc import Sequence

from .utils import FrozenAttrDict

__all__ = ("Options", "String", "Bool", "Integer", "Float", "Select", "Mapping")


class Options:
    """A declarative specification of exposed cluster options.

    Parameters
    ----------
    *fields : Field
        Zero or more configurable fields.
    handler : callable, optional
        A callable with the signature ``handler(options)`` or
        ``handler(options, user)``, where ``options`` is the validated dict of
        user options, and ``user`` is a ``User`` model for that user. Should
        return a dict of configuration overrides to forward to the cluster
        manager. If not provided, the default will return the options
        unchanged.

    Example
    -------

    Here we expose options for users to configure
    :data:`c.Backend.worker_cores` and
    :data:`c.Backend.worker_memory`. We set bounds on each resource to
    prevent users from requesting too large of a worker. The handler is used to
    convert the user specified memory from GiB to bytes (as expected by
    :data:`c.Backend.worker_memory`).

    .. code-block:: python

      from dask_gateway_server.options import Options, Integer, Float

      def options_handler(options):
          return {
              "worker_cores": options.worker_cores,
              "worker_memory": int(options.worker_memory * 2 ** 30)
          }

      c.Backend.DaskGateway.cluster_options = Options(
          Integer("worker_cores", default=1, min=1, max=4, label="Worker Cores"),
          Float("worker_memory", default=1, min=1, max=8, label="Worker Memory (GiB)"),
          handler=options_handler,
      )
    """

    def __init__(self, *fields, handler=None):
        for f in fields:
            if not isinstance(f, Field):
                raise TypeError(
                    "All fields must by instances of ``Field``, got %r"
                    % type(f).__name__
                )
        self.fields = fields
        self.handler = handler

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, handler):
        if handler is None:
            self._handler = None
        else:
            sig = inspect.signature(handler)
            if len(sig.parameters) == 1 and not any(
                a.kind == inspect.Parameter.VAR_POSITIONAL
                for a in sig.parameters.values()
            ):
                self._handler = lambda options, user: handler(options)
            else:
                self._handler = handler

    def get_specification(self):
        return [f.json_spec() for f in self.fields]

    def parse_options(self, request):
        if not isinstance(request, dict):
            raise TypeError("options must be a dict, got %r" % type(request).__name__)
        # Check for extra fields
        extra = set(request).difference(f.field for f in self.fields)
        if extra:
            raise ValueError("Unknown fields %r" % sorted(extra))
        # Validate options
        return {
            f.field: f.validate(request.get(f.field, f.get_default()))
            for f in self.fields
        }

    def transform_options(self, options):
        return {
            f.target: f.transform(options.get(f.field, f.get_default()))
            for f in self.fields
        }

    def get_configuration(self, options, user):
        options = self.transform_options(options)
        if self.handler is None:
            return options
        return self.handler(FrozenAttrDict(options), user)


_field_doc_template = """\
{description}

Parameters
----------
field : str
    The field name to use. Must be a valid Python variable name. This will
    be the keyword users use to set this field programmatically (e.g.
    ``"worker_cores"``).
{params}
label : str, optional
    A human readable label that will be used in GUI representations (e.g.
    ``"Worker Cores"``). If not provided, ``field`` will be used.
target : str, optional
    The target parameter to set in the processed options dict. Must be a
    valid Python variable name. If not provided, ``field`` will be used.
"""


def field_doc(description, params):
    def inner(cls):
        cls.__doc__ = _field_doc_template.format(
            description=description,
            params=textwrap.dedent(params.strip("\n")).strip("\n"),
        )
        return cls

    return inner


class Field:
    def __init__(self, field, default=None, target=None, label=None):
        self.field = field
        # Validate the default
        self.validate(default)
        self.default = default
        self.target = target or field
        self.label = label or field

    def json_spec(self):
        return {
            "field": self.field,
            "label": self.label,
            "default": self.default,
            "spec": self.json_type_spec(),
        }

    def json_type_spec(self):
        raise NotImplementedError

    def get_default(self):
        return self.default

    def validate(self, x):
        """Check that x is valid, and do any normalization.

        The output of this method must be serializable as json."""
        raise NotImplementedError

    def transform(self, x):
        """Transform a valid x into the desired output type.

        This may return any Python object."""
        return x


@field_doc(
    description="A string field.",
    params="""
    default : str, optional
        The default value. Default is the empty string (``""``).
    """,
)
class String(Field):
    def __init__(self, field, default="", label=None, target=None):
        super().__init__(field, default=default, label=label, target=target)

    def validate(self, x):
        if not isinstance(x, str):
            raise TypeError(f"{self.field} must be a string, got {x!r}")
        return x

    def json_type_spec(self):
        return {"type": "string"}


@field_doc(
    description="A boolean field.",
    params="""
    default : bool, optional
        The default value. Default is False.
    """,
)
class Bool(Field):
    def __init__(self, field, default=False, label=None, target=None):
        super().__init__(field, default=default, label=label, target=target)

    def validate(self, x):
        if not isinstance(x, bool):
            raise TypeError(f"{self.field} must be a bool, got {x!r}")
        return x

    def json_type_spec(self):
        return {"type": "bool"}


class Number(Field):
    def __init__(self, field, default=0, min=None, max=None, label=None, target=None):
        # Temporarily set to allow `validate` to work
        self.min = self.max = None
        if min is not None:
            self.min = self.validate(min)
        if max is not None:
            self.max = self.validate(max)
        super().__init__(field, default=default, label=label, target=target)

    def validate(self, x):
        if self.min is not None and x < self.min:
            raise ValueError(f"{self.field} must be >= {self.min:f}, got {x}")
        if self.max is not None and x > self.max:
            raise ValueError(f"{self.field} must be <= {self.max:f}, got {x}")
        return x


@field_doc(
    description="An integer field, with optional bounds.",
    params="""
    default : int, optional
        The default value. Default is 0.
    min : int, optional
        The minimum valid value (inclusive). Unbounded if not set.
    max : int, optional
        The maximum valid value (inclusive). Unbounded if not set.
    """,
)
class Integer(Number):
    def validate(self, x):
        if not isinstance(x, int):
            raise TypeError(f"{self.field} must be an integer, got {x!r}")
        return super().validate(x)

    def json_type_spec(self):
        return {"type": "int", "min": self.min, "max": self.max}


@field_doc(
    description="A float field, with optional bounds.",
    params="""
    default : float, optional
        The default value. Default is 0.
    min : float, optional
        The minimum valid value (inclusive). Unbounded if not set.
    max : float, optional
        The maximum valid value (inclusive). Unbounded if not set.
    """,
)
class Float(Number):
    def validate(self, x):
        if isinstance(x, int):
            x = float(x)
        if not isinstance(x, float):
            raise TypeError(f"{self.field} must be a float, got {x!r}")
        return super().validate(x)

    def json_type_spec(self):
        return {"type": "float", "min": self.min, "max": self.max}


@field_doc(
    description="A select field, allowing users to select between a few choices.",
    params="""
    options : list
        A list of valid options. Elements may be a tuple of ``(key, value)``,
        or just ``key`` (in which case the value is the same as the key).
        Values may be any Python object, keys must be strings.
    default : str, optional
        The key for the default option. Defaults to the first listed option.
    """,
)
class Select(Field):
    def __init__(self, field, options, default=None, label=None, target=None):
        if not isinstance(options, Sequence):
            raise TypeError("options must be a sequence")
        elif not len(options):
            raise ValueError("There must be at least one option")
        options_map = OrderedDict()
        for value in options:
            if isinstance(value, tuple):
                key, value = value
            else:
                key = value
            if not isinstance(key, str):
                raise TypeError("Select keys must be strings, got %r" % key)
            options_map[key] = value

        if default is None:
            default = list(options_map)[0]

        self.options = options_map
        super().__init__(field, default=default, label=label, target=target)

    def validate(self, x):
        if not isinstance(x, str):
            raise TypeError(f"{self.field} must be a string, got {x!r}")
        if x not in self.options:
            raise ValueError(f"{x!r} is not a valid option for {self.field}")
        return x

    def transform(self, x):
        self.validate(x)
        return self.options[x]

    def json_type_spec(self):
        return {"type": "select", "options": list(self.options)}


@field_doc(
    description="A mapping field.",
    params="""
    default : dict, optional
        The default value. Default is an empty dict (``{}``).
    """,
)
class Mapping(Field):
    def __init__(self, field, default=None, label=None, target=None):
        if default is None:
            default = {}
        super().__init__(field, default=default, label=label, target=target)

    def validate(self, x):
        if not isinstance(x, dict):
            raise TypeError(f"{self.field} must be a mapping, got {x!r}")
        return x

    def get_default(self):
        return copy.deepcopy(self.default)

    def json_type_spec(self):
        return {"type": "mapping"}



import asyncio
import collections

__all__ = ("Backoff", "WorkQueue", "WorkQueueClosed")


class Backoff:
    """Manages backoffs for items in the work queue.

    An items backoff is calculated as:

    ```
    min(base_delay * 2**n_failures, max_delay)
    ```
    """

    def __init__(self, base_delay=0.005, max_delay=300):
        self._failures = collections.defaultdict(int)
        self.base_delay = base_delay
        self.max_delay = max_delay

    def backoff(self, item):
        """Get the backoff time for an item (in seconds)"""
        exp = self._failures[item]
        self._failures[item] = exp + 1
        backoff = min(self.base_delay * 2**exp, self.max_delay)
        return backoff

    def failures(self, item):
        """Get the number of failures seen for an item"""
        return self._failures.get(item, 0)

    def reset(self, item):
        """Reset the backoff for an item"""
        self._failures.pop(item, None)


class WorkQueueClosed(Exception):
    """Indicates that the WorkQueue is closed."""

    pass


class WorkQueue:
    """An asynchronous work queue.

    This work queue is designed with the following design:
    - Supports multiple asynchronous producers/consumers
    - If a task is placed in the queue multiple times before being consumed,
      it will only be consumed once (duplicate puts are merged).
    - If a task is currently being processed, no other consumer may process
      this task (even if it's been re-added to the queue) until the task is
      marked as done using ``task_done``.
    - Tasks can be added after a delay with ``put_after``
    - Tasks can be requeued with backoff after a failure with ``put_backoff``.
      On success the backoff can be reset with ``reset_backoff``.
    """

    def __init__(self, backoff=None):
        self.backoff = backoff or Backoff()
        self._loop = asyncio.get_event_loop()
        self._waiting = collections.deque()
        self._queue = collections.deque()
        self._processing = set()
        self._dirty = set()
        self._timers = {}
        self.closed = False

    def is_empty(self):
        """True if there are no items queued"""
        return not self._dirty

    def _put(self, item):
        if item not in self._dirty:
            self._dirty.add(item)
            if item not in self._processing:
                self._queue.append(item)

    def _get(self):
        if self.closed:
            raise WorkQueueClosed
        item = self._queue.popleft()
        self._processing.add(item)
        self._dirty.discard(item)
        return item

    def _wakeup_next(self):
        while self._waiting:
            waiter = self._waiting.popleft()
            if not waiter.done():
                waiter.set_result(None)
                break

    def _wakeup_all(self):
        while self._waiting:
            waiter = self._waiting.popleft()
            if not waiter.done():
                waiter.set_result(None)

    def _put_delayed(self, item):
        self._timers.pop(item, None)
        self.put(item)

    def put(self, item):
        """Put an item in the queue"""
        self._put(item)
        self._wakeup_next()

    def put_after(self, item, delay):
        """Schedule an item to be put in the queue after a delay.

        If the item is already scheduled, it will be rescheduled only if the
        delay would enqueue it sooner than the existing schedule.
        """
        when = self._loop.time() + delay
        existing = self._timers.get(item, None)
        # Schedule if new or if sooner than existing schedule
        if existing is None or existing[1] > when:
            if existing is not None:
                existing[0].cancel()

            if delay > 0:
                self._timers[item] = (
                    self._loop.call_at(when, self._put_delayed, item),
                    when,
                )
            else:
                self._timers.pop(item, None)
                self.put(item)

    def put_backoff(self, item):
        """Schedule an item to be put in the queue after a backoff.

        If the item is already scheduled, it will be rescheduled only if the
        delay would enqueue it sooner than the existing schedule.
        """
        self.put_after(item, self.backoff.backoff(item))

    def failures(self, item):
        """Get the number of failures seen for this item"""
        return self.backoff.failures(item)

    def reset_backoff(self, item):
        """Reset the backoff for an item"""
        self.backoff.reset(item)

    async def get(self):
        """Get an item from the queue."""
        while not self._queue:
            if self.closed:
                raise WorkQueueClosed
            waiter = self._loop.create_future()
            self._waiting.append(waiter)
            try:
                await waiter
            except asyncio.CancelledError:
                try:
                    self._waiting.remove(waiter)
                except ValueError:
                    pass
                raise
        return self._get()

    def task_done(self, item):
        """Mark a task as done.

        This *must* be done before the item can be processed again.
        """
        self._processing.discard(item)
        if item in self._dirty:
            self._queue.append(item)
            self._wakeup_next()

    def close(self):
        """Close the queue.

        Future calls to ``WorkQueue.get`` will raise ``WorkQueueClosed``
        """
        self.closed = True
        self._wakeup_all()



import asyncio
import collections
import json
import logging
import signal
import sys
import time
import uuid
from base64 import b64encode

from aiohttp import web
from kubernetes_asyncio import client, config
from kubernetes_asyncio.client.rest import ApiException
from traitlets import Float, Integer, List, Unicode, validate
from traitlets.config import catch_config_error

from ... import __version__ as VERSION
from ...tls import new_keypair
from ...traitlets import Application
from ...utils import (
    AccessLogger,
    FrozenAttrDict,
    RateLimiter,
    TaskPool,
    normalize_address,
    run_main,
    timestamp,
)
from ...workqueue import Backoff, WorkQueue, WorkQueueClosed
from .backend import KubeBackendAndControllerMixin
from .utils import (
    Informer,
    RateLimitedClient,
    k8s_timestamp,
    merge_json_objects,
    parse_k8s_timestamp,
)


def get_container_status(pod, name):
    """Get the status of container ``name`` from a pod"""
    for cs in pod["status"].get("containerStatuses", ()):
        if cs["name"] == name:
            return cs
    return None


def get_container_state(pod, name):
    """Get the state of container ``name`` from a pod.

    Returns one of ``waiting, running, terminated, unknown``.
    """
    phase = pod["status"].get("phase", "Unknown")
    if phase == "Pending":
        return "waiting"
    cs = get_container_status(pod, name)
    if cs is not None:
        return next(iter(cs["state"]))
    return "unknown"


def get_cluster_key(obj):
    """Get the cluster key for a given k8s object"""
    try:
        namespace = obj["metadata"]["namespace"]
        name = obj["metadata"]["labels"]["gateway.dask.org/cluster"]
        return f"{namespace}.{name}"
    except KeyError:
        return None


class ClusterInfo:
    """Stores in-memory state about a given cluster.

    State can be reconstructed from k8s, this just provides fast access in the
    reconciler loop.
    """

    def __init__(self):
        self.all_pods = set()
        self.pending = set()
        self.running = set()
        self.succeeded = set()
        self.failed = set()
        self.set_expectations()

    def set_expectations(self, creates=0, deletes=0):
        """Indicate that this cluster expects to see ``creates``/``deletes`` pod
        operations before being woken up again.
        """
        self.creates = creates
        self.deletes = deletes
        self.timestamp = time.monotonic()

    def expectations_fulfilled(self):
        """Return true if all expectated pod operations have been seen"""
        return self.creates <= 0 and self.deletes <= 0

    def expectations_expired(self):
        """Return true if expected pod operations have timed out"""
        return time.monotonic() - self.timestamp > 600

    def should_trigger(self):
        """Return true if our expectations indicate the cluster should be
        reconciled again"""
        return self.expectations_fulfilled() or self.expectations_expired()

    def _update(self, kind, pod_name):
        if pod_name not in self.all_pods:
            self.creates -= 1
            self.all_pods.add(pod_name)

        for n in ("pending", "running", "succeeded", "failed"):
            s = getattr(self, n)
            if n == kind:
                change = pod_name not in s
                if change:
                    s.add(pod_name)
            else:
                s.discard(pod_name)
        return change

    def on_worker_pending(self, pod_name):
        return self._update("pending", pod_name)

    def on_worker_running(self, pod_name):
        return self._update("running", pod_name)

    def on_worker_succeeded(self, pod_name):
        return self._update("succeeded", pod_name)

    def on_worker_failed(self, pod_name):
        return self._update("failed", pod_name)

    def on_worker_deleted(self, pod_name):
        if pod_name in self.all_pods:
            self.deletes -= 1
            self.all_pods.discard(pod_name)
            for n in ("pending", "running", "succeeded", "failed"):
                getattr(self, n).discard(pod_name)


class KubeController(KubeBackendAndControllerMixin, Application):
    """Kubernetes controller for dask-gateway"""

    name = "dask-gateway-kube-controller"
    version = VERSION

    description = """Start dask-gateway kubernetes controller"""

    examples = """

    Start the controller with config file ``config.py``

        dask-gateway-kube-controller -f config.py
    """

    aliases = {
        "log-level": "KubeController.log_level",
        "f": "KubeController.config_file",
        "config": "KubeController.config_file",
    }

    config_file = Unicode(
        "dask_gateway_config.py", help="The config file to load", config=True
    )

    address = Unicode(
        ":8000", help="The address the server should listen at", config=True
    )

    @validate("address")
    def _validate_address(self, proposal):
        return normalize_address(proposal.value)

    api_url = Unicode(
        help="""
        The address that internal components (e.g. dask clusters)
        will use when contacting the gateway.
        """,
        config=True,
    )

    parallelism = Integer(
        20,
        help="""
        Number of handlers to use for reconciling k8s objects.
        """,
        config=True,
    )

    completed_cluster_cleanup_period = Float(
        600,
        help="""
        Time (in seconds) between cleanup tasks.

        This sets how frequently old cluster records are deleted from
        kubernetes.  This shouldn't be too small (to keep the overhead low),
        but should be smaller than ``completed_cluster_max_age`` (probably by an
        order of magnitude).
        """,
        config=True,
    )

    completed_cluster_max_age = Float(
        3600 * 24,
        help="""
        Max time (in seconds) to keep around records of completed clusters.

        Every ``completed_cluster_cleanup_period``, completed clusters older than
        ``completed_cluster_max_age`` are deleted from kubernetes.
        """,
        config=True,
    )

    backoff_base_delay = Float(
        0.1,
        help="""
        Base delay (in seconds) for backoff when retrying after failures.

        If an operation fails, it is retried after a backoff computed as:

        ```
        min(backoff_max_delay, backoff_base_delay * 2 ** num_failures)
        ```
        """,
        config=True,
    )

    backoff_max_delay = Float(
        300,
        help="""
        Max delay (in seconds) for backoff policy when retrying after failures.
        """,
        config=True,
    )

    k8s_api_rate_limit = Integer(
        50,
        help="""
        Limit on the average number of k8s api calls per second.
        """,
        config=True,
    )

    k8s_api_rate_limit_burst = Integer(
        100,
        help="""
        Limit on the maximum number of k8s api calls per second.
        """,
        config=True,
    )

    proxy_prefix = Unicode(
        "",
        help="""
        The path prefix the HTTP/HTTPS proxy should serve under.

        This prefix will be prepended to all routes registered with the proxy.
        """,
        config=True,
    )

    @validate("proxy_prefix")
    def _validate_proxy_prefix(self, proposal):
        prefix = proposal.value.strip("/")
        return f"/{prefix}" if prefix else prefix

    proxy_web_entrypoint = Unicode(
        "web",
        help="The traefik entrypoint name to use when creating ingressroutes",
        config=True,
    )

    proxy_tcp_entrypoint = Unicode(
        "tcp",
        help="The traefik entrypoint name to use when creating ingressroutetcps",
        config=True,
    )

    proxy_web_middlewares = List(
        help="A list of middlewares to apply to web routes added to the proxy.",
        config=True,
    )

    # Fail if the config file errors
    raise_config_file_errors = True

    def handle_shutdown_signal(self, sig):
        self.log.warning("Received signal %s, initiating shutdown...", sig.name)
        raise web.GracefulExit

    @catch_config_error
    def initialize(self, argv=None):
        super().initialize(argv)
        if self.subapp is not None:
            return

        # Setup logging
        self.log.propagate = False
        for name in ["aiohttp.access", "aiohttp.server"]:
            l = logging.getLogger(name)
            l.handlers[:] = []
            l.propagate = True
            l.parent = self.log
            l.setLevel(self.log_level)

        self.log.info("Starting %s - version %s", self.name, self.version)

        # Load configuration
        self.load_config_file(self.config_file)

        # Initialize aiohttp application
        self.app = web.Application(logger=self.log)
        self.app["controller"] = self

    def start(self):
        if self.subapp is not None:
            return self.subapp.start()

        try:
            run_main(self.main())
        except (KeyboardInterrupt, web.GracefulExit):
            pass

    async def main(self):
        try:
            try:
                await self.setup()
            except Exception:
                self.log.critical(
                    "Failed to start gateway, shutting down", exc_info=True
                )
                sys.exit(1)

            while True:
                await asyncio.sleep(3600)
        finally:
            try:
                await self.cleanup()
            except Exception:
                self.log.error("Error while shutting down:", exc_info=True)

    async def setup(self):
        # Register signal handlers
        loop = asyncio.get_event_loop()
        for s in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(s, self.handle_shutdown_signal, s)

        # Rate limiter for k8s api calls
        self.rate_limiter = RateLimiter(
            rate=self.k8s_api_rate_limit, burst=self.k8s_api_rate_limit_burst
        )

        # Initialize the kubernetes clients
        try:
            config.load_incluster_config()
        except config.ConfigException:
            await config.load_kube_config()
        self.api_client = client.ApiClient()
        self.core_client = RateLimitedClient(
            client.CoreV1Api(api_client=self.api_client), self.rate_limiter
        )
        self.custom_client = RateLimitedClient(
            client.CustomObjectsApi(api_client=self.api_client), self.rate_limiter
        )

        # Local state
        self.cluster_info = collections.defaultdict(ClusterInfo)
        self.stopped_clusters = {}

        # Initialize queue and informers
        self.queue = WorkQueue(
            backoff=Backoff(
                base_delay=self.backoff_base_delay, max_delay=self.backoff_max_delay
            )
        )
        endpoints_selector = (
            self.label_selector + ",app.kubernetes.io/component=dask-scheduler"
        )
        self.informers = {
            "cluster": Informer(
                parent=self,
                name="cluster",
                client=self.custom_client,
                method="list_cluster_custom_object",
                method_kwargs=dict(
                    group="gateway.dask.org",
                    version=self.crd_version,
                    plural="daskclusters",
                    label_selector=self.label_selector,
                ),
                on_update=self.on_cluster_update,
                on_delete=self.on_cluster_delete,
            ),
            "pod": Informer(
                parent=self,
                name="pod",
                client=self.core_client,
                method="list_pod_for_all_namespaces",
                method_kwargs=dict(label_selector=self.label_selector),
                on_update=self.on_pod_update,
                on_delete=self.on_pod_delete,
            ),
            "endpoints": Informer(
                parent=self,
                name="endpoints",
                client=self.core_client,
                method="list_endpoints_for_all_namespaces",
                method_kwargs=dict(label_selector=endpoints_selector),
                on_update=self.on_endpoints_update,
                on_delete=self.on_endpoints_delete,
            ),
        }
        await asyncio.wait(
            [asyncio.ensure_future(i.start()) for i in self.informers.values()]
        )
        self.log.debug("All informers started")

        # Initialize reconcilers
        self.reconcilers = [
            asyncio.ensure_future(self.reconciler_loop())
            for _ in range(self.parallelism)
        ]

        # Start background tasks
        self.task_pool = TaskPool()
        self.task_pool.spawn(self.cleanup_expired_cluster_records_loop())

        # Start the aiohttp application
        self.runner = web.AppRunner(
            self.app,
            handle_signals=False,
            access_log_class=AccessLogger,
            access_log=self.log,
        )
        await self.runner.setup()

        host, port = self.address.split(":")
        port = int(port)
        site = web.TCPSite(self.runner, host, port, shutdown_timeout=15.0, backlog=128)
        await site.start()
        self.log.info("%s started!", self.name)
        self.log.info("API listening at http://%s", self.address)

    async def cleanup(self):
        # Stop reconcilation workers
        if hasattr(self, "reconcilers"):
            self.queue.close()
            await asyncio.gather(*self.reconcilers, return_exceptions=True)

        if hasattr(self, "informers"):
            await asyncio.wait(
                [asyncio.ensure_future(i.stop()) for i in self.informers.values()]
            )

        # Stop background tasks
        if hasattr(self, "task_pool"):
            await self.task_pool.close()

        # Shutdown the client
        if hasattr(self, "api_client"):
            await self.api_client.rest_client.pool_manager.close()

        # Shutdown the aiohttp application
        if hasattr(self, "runner"):
            await self.runner.cleanup()

        self.log.info("Stopped successfully")

    async def delete_cluster(self, namespace, name):
        await self.custom_client.delete_namespaced_custom_object(
            "gateway.dask.org", self.crd_version, namespace, "daskclusters", name
        )

    async def cleanup_expired_cluster_records_loop(self):
        while True:
            try:
                cutoff = timestamp() - self.completed_cluster_max_age * 1000
                to_delete = [
                    k.split(".")
                    for k, stop_time in self.stopped_clusters.items()
                    if stop_time <= cutoff
                ]
                if to_delete:
                    self.log.info("Removing %d expired cluster records", len(to_delete))
                    tasks = (self.delete_cluster(ns, name) for ns, name in to_delete)
                    await asyncio.gather(*tasks, return_exceptions=True)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.error(
                    "Error while deleting expired cluster records", exc_info=exc
                )
            await asyncio.sleep(self.completed_cluster_cleanup_period)

    def on_pod_update(self, pod, old=None):
        cluster_key = get_cluster_key(pod)
        if cluster_key is None:
            return
        cluster_stopping = cluster_key in self.stopped_clusters

        component = pod["metadata"]["labels"]["app.kubernetes.io/component"]
        if component == "dask-scheduler":
            if (
                get_container_state(pod, "dask-scheduler") in ("running", "terminated")
                and not cluster_stopping
            ):
                self.queue.put(cluster_key)
        elif component == "dask-worker":
            pod_name = pod["metadata"]["name"]
            phase = pod["status"].get("phase", "Unknown")
            if phase == "Unknown":
                return
            cs = get_container_status(pod, component)
            if cs is None:
                return

            info = self.cluster_info[cluster_key]
            trigger = False
            if phase == "Pending":
                info.on_worker_pending(pod_name)
            elif phase == "Succeeded":
                info.on_worker_succeeded(pod_name)
            elif phase == "Failed":
                trigger = info.on_worker_failed(pod_name)
            else:
                kind = next(iter(cs["state"]))
                if kind == "terminated" and cs["state"][kind]["exitCode"] == 0:
                    info.on_worker_succeeded(pod_name)
                elif kind == "running":
                    info.on_worker_running(pod_name)

            if (trigger or info.should_trigger()) and not cluster_stopping:
                self.queue.put(cluster_key)

    def on_pod_delete(self, pod):
        cluster_key = get_cluster_key(pod)
        if cluster_key is None:
            return
        cluster_stopping = cluster_key in self.stopped_clusters

        component = pod["metadata"]["labels"]["app.kubernetes.io/component"]
        if component == "dask-scheduler" and not cluster_stopping:
            self.queue.put(cluster_key)
        elif component == "dask-worker":
            info = self.cluster_info[cluster_key]
            info.on_worker_deleted(pod["metadata"]["name"])
            if info.should_trigger() and not cluster_stopping:
                self.queue.put(cluster_key)

    def endpoints_all_ready(self, endpoints):
        subsets = endpoints.get("subsets", ())
        return subsets and not any(s.get("notReadyAddresses") for s in subsets)

    def on_endpoints_update(self, endpoints, old=None):
        cluster_key = get_cluster_key(endpoints)
        if cluster_key is None or cluster_key in self.stopped_clusters:
            return
        if self.endpoints_all_ready(endpoints):
            self.queue.put(cluster_key)

    def on_endpoints_delete(self, endpoints):
        pass

    def on_cluster_update(self, cluster, old=None):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]
        self.queue.put(f"{namespace}.{name}")

    def on_cluster_delete(self, cluster):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]
        self.log.debug("Cluster %s.%s deleted", namespace, name)
        self.stopped_clusters.pop(f"{namespace}.{name}", None)

    async def reconciler_loop(self):
        while True:
            try:
                name = await self.queue.get()
            except WorkQueueClosed:
                return

            self.log.debug("Reconciling cluster %s", name)
            try:
                requeue = await self.reconcile_cluster(name)
            except Exception:
                self.log.warning(
                    "Error while reconciling cluster %s", name, exc_info=True
                )
                self.queue.put_backoff(name)
            else:
                self.log.debug("Finished reconciling cluster %s", name)
                if requeue:
                    self.queue.put_backoff(name)
                else:
                    self.queue.reset_backoff(name)
            finally:
                self.queue.task_done(name)

    async def reconcile_cluster(self, cluster_key):
        cluster = self.informers["cluster"].get(cluster_key)
        if cluster is None:
            # Cluster already deleted, nothing to do
            return

        status_update, requeue = await self.handle_cluster(cluster)

        if status_update is not None and status_update != cluster.get("status"):
            namespace = cluster["metadata"]["namespace"]
            name = cluster["metadata"]["name"]
            await self.patch_cluster_status(namespace, name, status_update)
        return requeue

    async def handle_cluster(self, cluster):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]

        status = cluster.get("status", {})
        spec = cluster.get("spec")

        phase = status.get("phase", "Pending")

        if phase in {"Stopped", "Failed"}:
            if "completionTime" in status:
                stop_time = parse_k8s_timestamp(status["completionTime"])
            else:
                stop_time = timestamp()
            self.stopped_clusters[f"{namespace}.{name}"] = stop_time
            return None, False

        active = spec.get("active", True)

        if not active:
            self.log.info("Shutting down %s.%s", namespace, name)
            await self.cleanup_cluster_resources(status, namespace)
            status = status.copy()
            status["phase"] = "Stopped"
            status["completionTime"] = k8s_timestamp()
            return status, False

        if phase == "Pending":
            return await self.handle_pending_cluster(cluster)

        if phase == "Running":
            return await self.handle_running_cluster(cluster)

    async def handle_pending_cluster(self, cluster):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]

        status = cluster.get("status", {}).copy()
        status.setdefault("phase", "Pending")

        if not status.get("credentials"):
            secret_name = await self.create_secret_if_not_exists(cluster)
            status["credentials"] = secret_name

        sched_pod_name = status.get("schedulerPod")
        if not sched_pod_name:
            sched_pod_name, sched_pod = await self.create_scheduler_pod_if_not_exists(
                cluster
            )
            status["schedulerPod"] = sched_pod_name
        else:
            sched_pod = self.informers["pod"].get(f"{namespace}.{sched_pod_name}")

        if sched_pod is None:
            # Scheduler pod was created in an earlier run, but hasn't been seen
            # by the informers. Wait until it shows up to progress further.
            return status, False

        sched_state = get_container_state(sched_pod, "dask-scheduler")

        if sched_state == "running":
            service_name = status.get("service")
            if not status.get("service"):
                service_name = await self.create_service_if_not_exists(
                    cluster, sched_pod
                )
                status["service"] = service_name

            endpoints = self.informers["endpoints"].get(f"{namespace}.{service_name}")
            if endpoints is not None and self.endpoints_all_ready(endpoints):
                # Only add routes if endpoints exist, otherwise traefik gets grumpy
                # This also allows us to delay setting the phase to Running
                # until the route is likely to be added.
                if not status.get("ingressroute"):
                    route = await self.create_ingressroute_if_not_exists(
                        cluster, sched_pod
                    )
                    status["ingressroute"] = route

                if not status.get("ingressroutetcp"):
                    route = await self.create_ingressroutetcp_if_not_exists(
                        cluster, sched_pod
                    )
                    status["ingressroutetcp"] = route

                status["phase"] = "Running"
        elif sched_state == "terminated":
            self.log.info(
                "Scheduler for cluster %s.%s terminated, shutting down", namespace, name
            )
            await self.cleanup_cluster_resources(status, namespace)
            status["phase"] = "Failed"
            status["completionTime"] = k8s_timestamp()

        return status, False

    async def handle_running_cluster(self, cluster):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]
        cluster_key = f"{namespace}.{name}"
        requeue = False

        spec = cluster["spec"]
        status = cluster.get("status", {}).copy()

        sched_pod_name = status.get("schedulerPod")
        sched_pod = None
        sched_state = "terminated"
        if sched_pod_name:
            sched_pod = self.informers["pod"].get(f"{namespace}.{sched_pod_name}")
            if sched_pod is not None:
                sched_state = get_container_state(sched_pod, "dask-scheduler")

        if sched_state == "terminated":
            self.log.info(
                "Scheduler for cluster %s.%s terminated, shutting down", namespace, name
            )
            await self.cleanup_cluster_resources(status, namespace)
            status["phase"] = "Failed"
            status["completionTime"] = k8s_timestamp()

        elif sched_state == "running":
            info = self.cluster_info[cluster_key]
            if info.should_trigger():
                n_workers = len(info.running.union(info.pending))
                replicas = spec.get("replicas", 0)
                delta = replicas - n_workers
                if delta > 0:
                    requeue = await self.handle_scale_up(
                        cluster, sched_pod, info, replicas, delta
                    )
                else:
                    requeue = await self.handle_scale_down(
                        cluster, sched_pod, info, replicas, delta
                    )
        return status, requeue

    async def create_pod(self, namespace, pod, info=None):
        try:
            res = await self.core_client.create_namespaced_pod(namespace, pod)
            self.log.debug("Created pod %s/%s", namespace, res.metadata.name)
            return res
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            if info is not None:
                info.creates -= 1
            if isinstance(exc, ApiException):
                # Try to log a nicer message if an api exception
                try:
                    message = json.loads(exc.body)["message"]
                    self.log.warning(
                        "Failed to create pod in namespace %s - %s", namespace, message
                    )
                except Exception:
                    pass
                else:
                    raise
            self.log.warning(
                "Failed to create pod in namespace %s", namespace, exc_info=True
            )
            raise

    async def delete_pod(self, namespace, pod_name, info=None):
        try:
            await self.core_client.delete_namespaced_pod(pod_name, namespace)
            self.log.debug("Deleted pod %s/%s", namespace, pod_name)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            if info is not None:
                info.deletes -= 1
            if isinstance(exc, ApiException) and exc.code == 404:
                return
            self.log.warning(
                "Failed to delete pod %s/%s", namespace, pod_name, exc_info=True
            )
            raise

    async def batch_create_pods(self, info, namespace, pod, count):
        """Create pods in exponentially increasing batches.

        This helps prevent excessive failed api requests (which are sometimes
        expected, due to e.g. resource-quotas). It also reduces load on the k8s
        api server.
        """
        remaining = count
        batch_size = 1
        while remaining > 0:
            batch = [
                self.create_pod(namespace, pod, info)
                for _ in range(min(batch_size, remaining))
            ]
            res = await asyncio.gather(*batch, return_exceptions=True)
            remaining -= len(batch)
            if any(isinstance(r, Exception) for r in res):
                # Remaining creates will never be submitted
                info.creates -= remaining
                return True
            batch_size *= 2
        return False

    async def handle_scale_up(self, cluster, sched_pod, info, replicas, delta):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        config = FrozenAttrDict(cluster["spec"]["config"])

        pod = self.make_pod(namespace, name, config, is_worker=True)
        pod["metadata"]["ownerReferences"] = [
            {
                "apiVersion": "v1",
                "kind": "Pod",
                "name": sched_pod["metadata"]["name"],
                "uid": sched_pod["metadata"]["uid"],
            }
        ]
        to_delete = info.succeeded.union(info.failed)
        info.set_expectations(creates=delta, deletes=len(to_delete))
        self.log.info(
            "Cluster %s.%s scaled to %d - creating %d workers, deleting %d stopped workers",
            namespace,
            name,
            replicas,
            delta,
            len(to_delete),
        )
        failed = await self.batch_create_pods(info, namespace, pod, delta)
        res = await asyncio.gather(
            *(self.delete_pod(namespace, p, info) for p in to_delete),
            return_exceptions=True,
        )
        return failed or any(isinstance(r, Exception) for r in res)

    async def handle_scale_down(self, cluster, sched_pod, info, replicas, delta):
        namespace = cluster["metadata"]["namespace"]
        name = cluster["metadata"]["name"]

        if info.pending:
            pending = list(info.pending)[:delta]
        else:
            pending = []

        stopped = info.succeeded.union(info.failed)
        if pending or stopped:
            self.log.info(
                "Cluster %s.%s scaled to %d - deleting %d pending and %d stopped workers",
                namespace,
                name,
                replicas,
                len(pending),
                len(stopped),
            )
            pods = pending
            pods.extend(stopped)
            info.set_expectations(deletes=len(pods))
            res = await asyncio.gather(
                *(self.delete_pod(namespace, p, info) for p in pods),
                return_exceptions=True,
            )
            return any(isinstance(r, Exception) for r in res)
        return False

    async def patch_cluster_status(self, namespace, name, status):
        await self.custom_client.patch_namespaced_custom_object_status(
            "gateway.dask.org",
            self.crd_version,
            namespace,
            "daskclusters",
            name,
            [{"op": "add", "path": "/status", "value": status}],
        )

    async def cleanup_cluster_resources(self, status, namespace):
        sched_pod = status.get("schedulerPod")
        if sched_pod:
            try:
                await self.core_client.delete_namespaced_pod(sched_pod, namespace)
            except ApiException as exc:
                if exc.status != 404:
                    raise
        secret = status.get("credentials")
        if secret:
            try:
                await self.core_client.delete_namespaced_secret(secret, namespace)
            except ApiException as exc:
                if exc.status != 404:
                    raise

    async def create_secret_if_not_exists(self, cluster):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        secret = self.make_secret(name)
        secret["metadata"]["ownerReferences"] = [
            {
                "apiVersion": cluster["apiVersion"],
                "kind": cluster["kind"],
                "name": cluster["metadata"]["name"],
                "uid": cluster["metadata"]["uid"],
            }
        ]

        self.log.info("Creating new credentials for cluster %s.%s", namespace, name)
        try:
            await self.core_client.create_namespaced_secret(namespace, secret)
        except ApiException as exc:
            if exc.status != 409:
                raise

        return secret["metadata"]["name"]

    async def create_scheduler_pod_if_not_exists(self, cluster):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        config = FrozenAttrDict(cluster["spec"]["config"])
        pod = self.make_pod(namespace, name, config)
        pod["metadata"]["ownerReferences"] = [
            {
                "apiVersion": cluster["apiVersion"],
                "kind": cluster["kind"],
                "name": cluster["metadata"]["name"],
                "uid": cluster["metadata"]["uid"],
            }
        ]
        pod_name = pod["metadata"]["name"]

        self.log.info("Creating scheduler pod for cluster %s.%s", namespace, name)
        try:
            pod = await self.core_client.create_namespaced_pod(namespace, pod)
            pod = self.api_client.sanitize_for_serialization(pod)
        except ApiException as exc:
            if exc.status == 409:
                pod = None
            else:
                raise

        return pod_name, pod

    async def create_service_if_not_exists(self, cluster, sched_pod):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        service = self.make_service(name)
        service["metadata"]["ownerReferences"] = [
            {
                "apiVersion": "v1",
                "kind": "Pod",
                "name": sched_pod["metadata"]["name"],
                "uid": sched_pod["metadata"]["uid"],
            }
        ]

        self.log.info("Creating scheduler service for cluster %s.%s", namespace, name)
        try:
            await self.core_client.create_namespaced_service(namespace, service)
        except ApiException as exc:
            if exc.status != 409:
                raise

        return service["metadata"]["name"]

    async def create_ingressroute_if_not_exists(self, cluster, sched_pod):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        route = self.make_ingressroute(name, namespace)
        route["metadata"]["ownerReferences"] = [
            {
                "apiVersion": "v1",
                "kind": "Pod",
                "name": sched_pod["metadata"]["name"],
                "uid": sched_pod["metadata"]["uid"],
            }
        ]

        self.log.info(
            "Creating scheduler HTTP route for cluster %s.%s", namespace, name
        )
        try:
            await self.custom_client.create_namespaced_custom_object(
                "traefik.io", "v1alpha1", namespace, "ingressroutes", route
            )
        except ApiException as exc:
            if exc.status != 409:
                raise

        return route["metadata"]["name"]

    async def create_ingressroutetcp_if_not_exists(self, cluster, sched_pod):
        name = cluster["metadata"]["name"]
        namespace = cluster["metadata"]["namespace"]
        route = self.make_ingressroutetcp(name, namespace)
        route["metadata"]["ownerReferences"] = [
            {
                "apiVersion": "v1",
                "kind": "Pod",
                "name": sched_pod["metadata"]["name"],
                "uid": sched_pod["metadata"]["uid"],
            }
        ]

        self.log.info("Creating scheduler TCP route for cluster %s.%s", namespace, name)
        try:
            await self.custom_client.create_namespaced_custom_object(
                "traefik.io", "v1alpha1", namespace, "ingressroutetcps", route
            )
        except ApiException as exc:
            if exc.status != 409:
                raise

        return route["metadata"]["name"]

    def get_scheduler_command(self, namespace, cluster_name, config):
        return config.scheduler_cmd + [
            "--protocol",
            "tls",
            "--host",
            "",
            "--port",
            "8786",
            "--dashboard-address",
            ":8787",
            "--dg-api-address",
            ":8788",
            "--preload",
            "dask_gateway.scheduler_preload",
            "--dg-heartbeat-period",
            "0",
            "--dg-adaptive-period",
            str(config.adaptive_period),
            "--dg-idle-timeout",
            str(config.idle_timeout),
        ]

    def get_worker_command(self, namespace, cluster_name, config):
        service_name = self.make_service_name(cluster_name)
        return config.worker_cmd + [
            f"tls://{service_name}.{namespace}:8786",
            "--dashboard-address",
            ":8787",
            "--name",
            "$(DASK_GATEWAY_WORKER_NAME)",
            "--nthreads",
            str(config.worker_threads),
            "--memory-limit",
            str(config.worker_memory_limit),
        ]

    def get_env(self, namespace, cluster_name, config):
        env = dict(config.environment)
        env.update(
            {
                "DASK_GATEWAY_API_URL": self.api_url,
                "DASK_GATEWAY_CLUSTER_NAME": f"{namespace}.{cluster_name}",
                "DASK_GATEWAY_API_TOKEN": "/etc/dask-credentials/api-token",
                "DASK_DISTRIBUTED__COMM__TLS__CA_FILE": "/etc/dask-credentials/dask.crt",
                "DASK_DISTRIBUTED__COMM__TLS__SCHEDULER__CERT": "/etc/dask-credentials/dask.crt",
                "DASK_DISTRIBUTED__COMM__TLS__SCHEDULER__KEY": "/etc/dask-credentials/dask.pem",
                "DASK_DISTRIBUTED__COMM__TLS__WORKER__CERT": "/etc/dask-credentials/dask.crt",
                "DASK_DISTRIBUTED__COMM__TLS__WORKER__KEY": "/etc/dask-credentials/dask.pem",
            }
        )
        return [{"name": k, "value": v} for k, v in env.items()]

    def get_labels(self, cluster_name, component=None):
        labels = self.common_labels.copy()
        labels.update(
            {
                "gateway.dask.org/instance": self.gateway_instance,
                "gateway.dask.org/cluster": cluster_name,
            }
        )
        if component:
            labels["app.kubernetes.io/component"] = component
        return labels

    def make_pod(self, namespace, cluster_name, config, is_worker=False):
        env = self.get_env(namespace, cluster_name, config)

        if is_worker:
            container_name = "dask-worker"
            mem_req = config.worker_memory
            mem_lim = config.worker_memory_limit
            cpu_req = config.worker_cores
            cpu_lim = config.worker_cores_limit
            cmd = self.get_worker_command(namespace, cluster_name, config)
            env.append(
                {
                    "name": "DASK_GATEWAY_WORKER_NAME",
                    "valueFrom": {"fieldRef": {"fieldPath": "metadata.name"}},
                }
            )
            extra_pod_config = config.worker_extra_pod_config
            extra_container_config = config.worker_extra_container_config
            extra_pod_annotations = config.worker_extra_pod_annotations
            extra_pod_labels = config.worker_extra_pod_labels
            probes = {}
        else:
            container_name = "dask-scheduler"
            mem_req = config.scheduler_memory
            mem_lim = config.scheduler_memory_limit
            cpu_req = config.scheduler_cores
            cpu_lim = config.scheduler_cores_limit
            cmd = self.get_scheduler_command(namespace, cluster_name, config)
            extra_pod_config = config.scheduler_extra_pod_config
            extra_container_config = config.scheduler_extra_container_config
            extra_pod_annotations = config.scheduler_extra_pod_annotations
            extra_pod_labels = config.scheduler_extra_pod_labels
            # TODO: make this configurable. If supported, we should use a
            # startupProbe here instead.
            probes = {
                "readinessProbe": {
                    "httpGet": {"port": 8788, "path": "/api/health"},
                    "periodSeconds": 5,
                    "failureThreshold": 3,
                }
            }

        volume = {
            "name": "dask-credentials",
            "secret": {"secretName": self.make_secret_name(cluster_name)},
        }

        container = {
            "name": container_name,
            "image": config.image,
            "args": cmd,
            "env": env,
            "resources": {
                "requests": {"cpu": f"{cpu_req:.3f}", "memory": str(mem_req)},
                "limits": {"cpu": f"{cpu_lim:.3f}", "memory": str(mem_lim)},
            },
            "volumeMounts": [
                {
                    "name": "dask-credentials",
                    "mountPath": "/etc/dask-credentials/",
                    "readOnly": True,
                }
            ],
            "ports": [
                {"name": "scheduler", "containerPort": 8786},
                {"name": "dashboard", "containerPort": 8787},
                {"name": "api", "containerPort": 8788},
            ],
        }
        if config.image_pull_policy:
            container["imagePullPolicy"] = config.image_pull_policy

        container.update(probes)

        if extra_container_config:
            container = merge_json_objects(container, extra_container_config)

        annotations = self.common_annotations.copy()
        annotations.update(extra_pod_annotations)

        labels = self.get_labels(cluster_name, container_name)
        labels.update(extra_pod_labels)

        pod = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"labels": labels, "annotations": annotations},
            "spec": {
                "containers": [container],
                "volumes": [volume],
                "restartPolicy": "OnFailure" if is_worker else "Never",
                "automountServiceAccountToken": False,
            },
        }

        if config.image_pull_secrets:
            pod["spec"]["imagePullSecrets"] = config.image_pull_secrets

        if extra_pod_config:
            pod["spec"] = merge_json_objects(pod["spec"], extra_pod_config)

        if is_worker:
            pod["metadata"]["generateName"] = f"dask-worker-{cluster_name}-"
        else:
            pod["metadata"]["name"] = f"dask-scheduler-{cluster_name}"

        return pod

    def make_secret_name(self, cluster_name):
        return f"dask-credentials-{cluster_name}"

    def make_secret(self, cluster_name):
        api_token = uuid.uuid4().hex
        tls_cert, tls_key = new_keypair(cluster_name)

        labels = self.get_labels(cluster_name, "credentials")

        return {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {
                "labels": labels,
                "annotations": self.common_annotations,
                "name": self.make_secret_name(cluster_name),
            },
            "data": {
                "dask.crt": b64encode(tls_cert).decode(),
                "dask.pem": b64encode(tls_key).decode(),
                "api-token": b64encode(api_token.encode()).decode(),
            },
        }

    def make_service_name(self, cluster_name):
        return f"dask-{cluster_name}"

    def make_service(self, cluster_name):
        return {
            "metadata": {
                "labels": self.get_labels(cluster_name, "dask-scheduler"),
                "annotations": self.common_annotations,
                "name": self.make_service_name(cluster_name),
            },
            "spec": {
                "clusterIP": "None",
                "selector": {
                    "gateway.dask.org/cluster": cluster_name,
                    "gateway.dask.org/instance": self.gateway_instance,
                    "app.kubernetes.io/component": "dask-scheduler",
                },
                "ports": [
                    {"name": "scheduler", "port": 8786, "target_port": "scheduler"},
                    {"name": "dashboard", "port": 8787, "target_port": "dashboard"},
                    {"name": "api", "port": 8788, "target_port": "api"},
                ],
            },
        }

    def make_ingressroute(self, cluster_name, namespace):
        route = f"{self.proxy_prefix}/clusters/{namespace}.{cluster_name}/"
        return {
            "apiVersion": "traefik.io/v1alpha1",
            "kind": "IngressRoute",
            "metadata": {
                "labels": self.get_labels(cluster_name, "dask-scheduler"),
                "annotations": self.common_annotations,
                "name": f"dask-{cluster_name}",
            },
            "spec": {
                "entryPoints": [self.proxy_web_entrypoint],
                "routes": [
                    {
                        "kind": "Rule",
                        "match": f"PathPrefix(`{route}`)",
                        "services": [
                            {
                                "name": self.make_service_name(cluster_name),
                                "namespace": namespace,
                                "port": 8787,
                            }
                        ],
                        "middlewares": self.proxy_web_middlewares,
                    }
                ],
            },
        }

    def make_ingressroutetcp(self, cluster_name, namespace):
        return {
            "apiVersion": "traefik.io/v1alpha1",
            "kind": "IngressRouteTCP",
            "metadata": {
                "labels": self.get_labels(cluster_name, "dask-scheduler"),
                "annotations": self.common_annotations,
                "name": f"dask-{cluster_name}",
            },
            "spec": {
                "entryPoints": [self.proxy_tcp_entrypoint],
                "routes": [
                    {
                        "match": f"HostSNI(`daskgateway-{namespace}.{cluster_name}`)",
                        "services": [
                            {
                                "name": self.make_service_name(cluster_name),
                                "namespace": namespace,
                                "port": 8786,
                            }
                        ],
                    }
                ],
                "tls": {"passthrough": True},
            },
        }


main = KubeController.launch_instance

if __name__ == "__main__":
    main()



import asyncio
import socket
import time
import weakref
from collections import OrderedDict
from collections.abc import Mapping
from inspect import isawaitable
from keyword import iskeyword

import aiohttp.abc
from colorlog import ColoredFormatter


def timestamp():
    """An integer timestamp represented as milliseconds since the epoch UTC"""
    return int(time.time() * 1000)


class TaskPool:
    def __init__(self):
        self.tasks = weakref.WeakSet()
        self.closed = False

    def spawn(self, task):
        out = asyncio.ensure_future(task)
        self.tasks.add(out)
        return out

    async def close(self):
        # Stop all tasks
        self.closed = True
        for task in self.tasks:
            task.cancel()
        await asyncio.gather(*self.tasks, return_exceptions=True)
        self.tasks.clear()


class CancelGroup:
    def __init__(self):
        self.tasks = set()
        self.cancelled = False

    def cancellable(self):
        return _cancel_context(self)

    async def cancel(self):
        if self.cancelled:
            raise asyncio.CancelledError
        else:
            self.cancelled = True
            tasks = [t._task for t in self.tasks if t._task is not None]
            for t in tasks:
                t.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            self.tasks.clear()

    def _register(self, it):
        self.tasks.add(it)

    def _unregister(self, it):
        self.tasks.discard(it)


class _cancel_context:
    def __init__(self, context):
        self.context = context
        self._task = None

    async def __aenter__(self):
        if self.context.cancelled:
            raise asyncio.CancelledError
        try:
            self._task = asyncio.current_task()
        except AttributeError:
            self._task = asyncio.Task.current_task()
        self.context._register(self)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.context._unregister(self)
        self._task = None


class RateLimiter:
    """A token-bucket based rate limiter.

    The bucket starts out with ``burst`` tokens, and is replenished at ``rate``
    tokens per-second. Calls to ``acquire`` take a token, possibly blocking
    until one is available.
    """

    def __init__(self, rate=10, burst=100):
        self.rate = float(rate)
        self.burst = float(burst)
        self._tokens = float(burst)
        self._max_elapsed = burst / rate

        now = time.monotonic()
        self._last = now
        self._last_event = now

    def _delay(self):
        now = time.monotonic()
        elapsed = min(now - self._last, self._max_elapsed)
        tokens = min(self._tokens + elapsed * self.rate, self.burst)

        tokens -= 1
        if tokens < 0:
            delay = -tokens / self.rate
        else:
            delay = 0

        self._tokens = tokens
        self._last = now
        self._last_event = now + delay
        return delay

    async def acquire(self):
        """Acquire a rate-limited token."""
        delay = self._delay()
        if delay:
            await asyncio.sleep(delay)


def random_port():
    """Get a single random port."""
    with socket.socket() as sock:
        sock.bind(("", 0))
        return sock.getsockname()[1]


def normalize_address(address, resolve_host=False):
    try:
        host, port = address.split(":")
        port = int(port)
        if resolve_host and host in {"", "0.0.0.0"}:
            host = socket.gethostname()
        if port == 0:
            port = random_port()
    except Exception:
        raise ValueError(f"Invalid address `{address}`, should be of form `host:port`")

    return f"{host}:{port}"


async def cancel_task(task):
    if task.done():
        return
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


def format_bytes(nbytes):
    """Format ``nbytes`` as a human-readable string with units"""
    if nbytes > 2**50:
        return "%0.2f PiB" % (nbytes / 2**50)
    if nbytes > 2**40:
        return "%0.2f TiB" % (nbytes / 2**40)
    if nbytes > 2**30:
        return "%0.2f GiB" % (nbytes / 2**30)
    if nbytes > 2**20:
        return "%0.2f MiB" % (nbytes / 2**20)
    if nbytes > 2**10:
        return "%0.2f KiB" % (nbytes / 2**10)
    return "%d B" % nbytes


def classname(cls):
    """Return the full qualified name of a class"""
    mod = cls.__module__
    return cls.__name__ if mod is None else f"{mod}.{cls.__name__}"


class LogFormatter(ColoredFormatter):
    def __init__(self, fmt=None, datefmt=None, style=None):
        super().__init__(
            fmt=fmt,
            datefmt=datefmt,
            style=style,
            reset=False,
            log_colors={
                "DEBUG": "blue",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )


class AccessLogger(aiohttp.abc.AbstractAccessLogger):
    def log(self, request, response, time):
        if response.status >= 500:
            level = self.logger.error
        elif response.status >= 400:
            level = self.logger.warning
        elif request.path.endswith("/api/health"):
            level = self.logger.debug
        else:
            level = self.logger.info
        level(
            "%d %s %s %.3fms",
            response.status,
            request.method,
            request.path_qs,
            time * 1000,
        )


async def awaitable(obj):
    if isawaitable(obj):
        return await obj
    return obj


class FrozenAttrDict(Mapping):
    """A dict that also allows attribute access for keys"""

    __slots__ = ("_mapping",)

    def __init__(self, mapping):
        self._mapping = mapping

    def __getattr__(self, k):
        if k in self._mapping:
            return self._mapping[k]
        raise AttributeError(k)

    def __getitem__(self, k):
        return self._mapping[k]

    def __iter__(self):
        return iter(self._mapping)

    def __len__(self):
        return len(self._mapping)

    def __dir__(self):
        out = set(dir(type(self)))
        out.update(k for k in self if k.isidentifier() and not iskeyword(k))
        return list(out)


class LRUCache:
    """A LRU cache"""

    def __init__(self, max_size):
        self.cache = OrderedDict()
        self.max_size = max_size

    def get(self, key):
        """Get an item from the cache. Returns None if not present"""
        try:
            self.cache.move_to_end(key)
            return self.cache[key]
        except KeyError:
            return None

    def put(self, key, value):
        """Add an item to the cache"""
        self.cache[key] = value
        if len(self.cache) > self.max_size:
            self.cache.popitem(False)

    def discard(self, key):
        """Remove an item from the cache. No-op if not present."""
        try:
            del self.cache[key]
        except KeyError:
            pass


class Flag:
    """A simpler version of asyncio.Event"""

    def __init__(self):
        self._future = asyncio.get_running_loop().create_future()

    def set(self):
        if not self._future.done():
            self._future.set_result(None)

    def __await__(self):
        return asyncio.shield(self._future).__await__()

    def is_set(self):
        return self._future.done()


def run_main(main):
    """Main entrypoint for asyncio tasks.

    This differs from `asyncio.run` in one key way - the main task is cancelled
    *first*, then any outstanding tasks are cancelled (and logged, remaining
    tasks are indicative of bugs/failures).
    """
    if asyncio._get_running_loop() is not None:
        raise RuntimeError("Cannot be called from inside a running event loop")

    main_task = None
    loop = asyncio.new_event_loop()
    try:
        asyncio.set_event_loop(loop)
        main_task = loop.create_task(main)
        return loop.run_until_complete(main_task)
    finally:
        try:
            if main_task is not None:
                loop.run_until_complete(cancel_task(main_task))
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.set_event_loop(None)
            loop.close()


def _cancel_all_tasks(loop):
    to_cancel = asyncio.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )



import asyncio
import logging
import os
import signal
import sys

from aiohttp import web
from traitlets import Bool, List, Unicode, default, validate
from traitlets.config import catch_config_error

from . import __version__ as VERSION
from .auth import Authenticator
from .backends import Backend
from .routes import default_routes
from .traitlets import Application, Type
from .utils import AccessLogger, classname, normalize_address, run_main


class GenerateConfig(Application):
    """Generate and write a default configuration file"""

    name = "dask-gateway-server generate-config"
    version = VERSION
    description = "Generate and write a default configuration file"

    examples = """

        dask-gateway-server generate-config
    """

    output = Unicode(
        "dask_gateway_config.py", help="The path to write the config file", config=True
    )

    force = Bool(False, help="If true, will overwrite file if it exists.", config=True)

    aliases = {"output": "GenerateConfig.output"}

    flags = {
        "force": (
            {"GenerateConfig": {"force": True}},
            "Overwrite config file if it exists",
        )
    }

    def start(self):
        config_file_dir = os.path.dirname(os.path.abspath(self.output))
        if not os.path.isdir(config_file_dir):
            self.exit(
                "%r does not exist. The destination directory must exist "
                "before generating config file." % config_file_dir
            )
        if os.path.exists(self.output) and not self.force:
            self.exit("Config file already exists, use `--force` to overwrite")

        config_text = DaskGateway.instance().generate_config_file()
        if isinstance(config_text, bytes):
            config_text = config_text.decode("utf8")
        print("Writing default config to: %s" % self.output)
        with open(self.output, mode="w") as f:
            f.write(config_text)


class DaskGateway(Application):
    """A gateway for managing dask clusters across multiple users"""

    name = "dask-gateway-server"
    version = VERSION

    description = """Start a Dask Gateway server"""

    examples = """

    Start the server with config file ``config.py``

        dask-gateway-server -f config.py
    """

    subcommands = {
        "generate-config": (
            "dask_gateway_server.app.GenerateConfig",
            "Generate a default config file",
        ),
        "proxy": ("dask_gateway_server.proxy.core.ProxyApp", "Start the proxy"),
        "kube-controller": (
            "dask_gateway_server.backends.kubernetes.controller.KubeController",
            "Start the k8s controller",
        ),
    }

    aliases = {
        "log-level": "DaskGateway.log_level",
        "f": "DaskGateway.config_file",
        "config": "DaskGateway.config_file",
    }

    config_file = Unicode(
        "dask_gateway_config.py", help="The config file to load", config=True
    )

    # Fail if the config file errors
    raise_config_file_errors = True

    authenticator_class = Type(
        "dask_gateway_server.auth.SimpleAuthenticator",
        klass="dask_gateway_server.auth.Authenticator",
        help="The gateway authenticator class to use",
        config=True,
    )

    backend_class = Type(
        klass="dask_gateway_server.backends.Backend",
        help="The gateway backend class to use",
        config=True,
    )

    @default("backend_class")
    def _default_backend_class(self):
        # Separate the default out to avoid traitlets auto-importing it,
        # even if it's unused :(.
        return "dask_gateway_server.backends.local.UnsafeLocalBackend"

    address = Unicode(
        help="""
        The address the private api server should *listen* at.

        Should be of the form ``{hostname}:{port}``

        Where:

        - ``hostname`` sets the hostname to *listen* at. Set to ``""`` or
          ``"0.0.0.0"`` to listen on all interfaces.
        - ``port`` sets the port to *listen* at.

        Defaults to ``127.0.0.1:0``.
        """,
        config=True,
    )

    @default("address")
    def _default_address(self):
        return normalize_address("127.0.0.1:0")

    @validate("address")
    def _validate_address(self, proposal):
        return normalize_address(proposal.value)

    classes = List([Backend, Authenticator])

    @catch_config_error
    def initialize(self, argv=None):
        super().initialize(argv)
        if self.subapp is not None:
            return

        # Setup logging
        self.log.propagate = False
        for name in ["aiohttp.access", "aiohttp.server"]:
            l = logging.getLogger(name)
            l.handlers[:] = []
            l.propagate = True
            l.parent = self.log
            l.setLevel(self.log_level)

        self.log.info("Starting dask-gateway-server - version %s", VERSION)

        # Load configuration
        self.load_config_file(self.config_file)

        # Initialize authenticator and backend
        self.log.info("Authenticator: %r", classname(self.authenticator_class))
        self.authenticator = self.authenticator_class(parent=self, log=self.log)
        self.log.info("Backend: %r", classname(self.backend_class))
        self.backend = self.backend_class(parent=self, log=self.log)

        # Initialize aiohttp application
        self.app = web.Application(logger=self.log)
        self.app.add_routes(default_routes)
        self.app["gateway"] = self
        self.app["backend"] = self.backend
        self.app["authenticator"] = self.authenticator
        self.app["log"] = self.log

    def handle_shutdown_signal(self, sig):
        self.log.warning("Received signal %s, initiating shutdown...", sig.name)
        raise web.GracefulExit

    async def setup(self):
        # Register signal handlers
        loop = asyncio.get_event_loop()
        for s in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(s, self.handle_shutdown_signal, s)

        # Start the authenticator
        await self.authenticator.setup(self.app)

        # Start the backend
        await self.backend.setup(self.app)

        # Start the aiohttp application
        self.runner = web.AppRunner(
            self.app,
            handle_signals=False,
            access_log_class=AccessLogger,
            access_log=self.log,
        )
        await self.runner.setup()

        host, port = self.address.split(":")
        port = int(port)
        site = web.TCPSite(self.runner, host, port, shutdown_timeout=15.0, backlog=128)
        await site.start()
        self.log.info("Dask-Gateway server started")
        self.log.info("- Private API server listening at http://%s", self.address)

    async def cleanup(self):
        # Shutdown the aiohttp application
        if hasattr(self, "runner"):
            await self.runner.cleanup()

        # Shutdown backend
        if hasattr(self, "backend"):
            await self.backend.cleanup()

        # Shutdown authenticator
        if hasattr(self, "authenticator"):
            await self.authenticator.cleanup()

        self.log.info("Stopped successfully")

    def start(self):
        if self.subapp is not None:
            return self.subapp.start()

        try:
            run_main(self.main())
        except (KeyboardInterrupt, web.GracefulExit):
            pass

    async def main(self):
        try:
            try:
                await self.setup()
            except Exception:
                self.log.critical(
                    "Failed to start gateway, shutting down", exc_info=True
                )
                sys.exit(1)

            while True:
                await asyncio.sleep(3600)
        finally:
            try:
                await self.cleanup()
            except Exception:
                self.log.error("Error while shutting down:", exc_info=True)

    async def health(self):
        # TODO: add runtime checks here
        return {"status": "pass"}

    def version_info(self):
        """Handles the /api/version route"""
        return {
            "version": VERSION,
        }


main = DaskGateway.launch_instance



import asyncio
import json
import os
import pwd
import shutil

from traitlets import Unicode, default

from ..base import ClusterConfig
from ..db_base import DBBackendBase

__all__ = ("JobQueueClusterConfig", "JobQueueBackend")


class JobQueueClusterConfig(ClusterConfig):
    worker_setup = Unicode(
        "", help="Script to run before dask worker starts.", config=True
    )

    scheduler_setup = Unicode(
        "", help="Script to run before dask scheduler starts.", config=True
    )

    staging_directory = Unicode(
        "{home}/.dask-gateway/",
        help="""
        The staging directory for storing files before the job starts.

        A subdirectory will be created for each new cluster which will store
        temporary files for that cluster. On cluster shutdown the subdirectory
        will be removed.

        This field can be a template, which receives the following fields:

        - home (the user's home directory)
        - username (the user's name)
        """,
        config=True,
    )


class JobQueueBackend(DBBackendBase):
    """A base cluster manager for deploying Dask on a jobqueue cluster."""

    dask_gateway_jobqueue_launcher = Unicode(
        help="The path to the dask-gateway-jobqueue-launcher executable", config=True
    )

    @default("dask_gateway_jobqueue_launcher")
    def _default_launcher_path(self):
        return (
            shutil.which("dask-gateway-jobqueue-launcher")
            or "dask-gateway-jobqueue-launcher"
        )

    submit_command = Unicode(help="The path to the job submit command", config=True)

    cancel_command = Unicode(help="The path to the job cancel command", config=True)

    status_command = Unicode(help="The path to the job status command", config=True)

    def get_submit_cmd_env_stdin(self, cluster, worker=None):
        raise NotImplementedError

    def get_stop_cmd_env(self, job_id):
        raise NotImplementedError

    def get_status_cmd_env(self, job_ids):
        raise NotImplementedError

    def parse_job_id(self, stdout):
        raise NotImplementedError

    def parse_job_states(self, stdout):
        raise NotImplementedError

    def get_staging_directory(self, cluster):
        staging_dir = cluster.config.staging_directory.format(
            home=pwd.getpwnam(cluster.username).pw_dir, username=cluster.username
        )
        return os.path.join(staging_dir, cluster.name)

    def get_tls_paths(self, cluster):
        """Get the absolute paths to the tls cert and key files."""
        staging_dir = self.get_staging_directory(cluster)
        cert_path = os.path.join(staging_dir, "dask.crt")
        key_path = os.path.join(staging_dir, "dask.pem")
        return cert_path, key_path

    async def do_as_user(self, user, action, **kwargs):
        cmd = ["sudo", "-nHu", user, self.dask_gateway_jobqueue_launcher]
        kwargs["action"] = action
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            env={},
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await proc.communicate(json.dumps(kwargs).encode("utf8"))
        stdout = stdout.decode("utf8", "replace")
        stderr = stderr.decode("utf8", "replace")

        if proc.returncode != 0:
            raise Exception(
                "Error running `dask-gateway-jobqueue-launcher`\n"
                "  returncode: %d\n"
                "  stdout: %s\n"
                "  stderr: %s" % (proc.returncode, stdout, stderr)
            )
        result = json.loads(stdout)
        if not result["ok"]:
            raise Exception(result["error"])
        return result["returncode"], result["stdout"], result["stderr"]

    async def start_job(self, username, cmd, env, stdin, staging_dir=None, files=None):
        code, stdout, stderr = await self.do_as_user(
            user=username,
            action="start",
            cmd=cmd,
            env=env,
            stdin=stdin,
            staging_dir=staging_dir,
            files=files,
        )
        if code != 0:
            raise Exception(
                (
                    "Failed to submit job to batch system\n"
                    "  exit_code: %d\n"
                    "  stdout: %s\n"
                    "  stderr: %s"
                )
                % (code, stdout, stderr)
            )
        return self.parse_job_id(stdout)

    async def stop_job(self, username, job_id, staging_dir=None):
        cmd, env = self.get_stop_cmd_env(job_id)

        code, stdout, stderr = await self.do_as_user(
            user=username, action="stop", cmd=cmd, env=env, staging_dir=staging_dir
        )
        if code != 0 and "Job has finished" not in stderr:
            raise Exception("Failed to stop job_id %s" % job_id)

    async def check_jobs(self, job_ids):
        if not job_ids:
            return {}
        self.log.debug("Checking status of %d jobs", len(job_ids))
        cmd, env = self.get_status_cmd_env(job_ids)
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            env=env,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await proc.communicate()
        stdout = stdout.decode("utf8", "replace")
        if proc.returncode != 0:
            stderr = stderr.decode("utf8", "replace")
            self.log.warning(
                "Job status check failed with returncode %d, stderr: %s",
                proc.returncode,
                stderr,
            )
            raise Exception("Job status check failed")
        return self.parse_job_states(stdout)

    async def do_start_cluster(self, cluster):
        cmd, env, stdin = self.get_submit_cmd_env_stdin(cluster)
        staging_dir = self.get_staging_directory(cluster)
        files = {
            "dask.pem": cluster.tls_key.decode("utf8"),
            "dask.crt": cluster.tls_cert.decode("utf8"),
        }
        job_id = await self.start_job(
            cluster.username, cmd, env, stdin, staging_dir=staging_dir, files=files
        )
        self.log.info("Job %s submitted for cluster %s", job_id, cluster.name)
        yield {"job_id": job_id, "staging_dir": staging_dir}

    async def do_stop_cluster(self, cluster):
        job_id = cluster.state.get("job_id")
        if job_id is not None:
            staging_dir = cluster.state["staging_dir"]
            await self.stop_job(cluster.username, job_id, staging_dir=staging_dir)

    async def do_start_worker(self, worker):
        cmd, env, stdin = self.get_submit_cmd_env_stdin(worker.cluster, worker)
        job_id = await self.start_job(worker.cluster.username, cmd, env, stdin)
        self.log.info("Job %s submitted for worker %s", job_id, worker.name)
        yield {"job_id": job_id}

    async def do_stop_worker(self, worker):
        job_id = worker.state.get("job_id")
        if job_id is not None:
            await self.stop_job(worker.cluster.username, job_id)

    async def _do_check(self, objs):
        id_map = {}
        for x in objs:
            job_id = x.state.get("job_id")
            if job_id is not None:
                id_map[x.name] = job_id
        states = await self.check_jobs(list(id_map.values()))
        return [states.get(id_map.get(x.name), False) for x in objs]

    async def do_check_clusters(self, clusters):
        return await self._do_check(clusters)

    async def do_check_workers(self, workers):
        return await self._do_check(workers)
